#!/usr/bin/env python

import pickle, sys
from spt3g import core

f = core.G3Frame(core.G3FrameType.Scan)
f['Five'] = 5

b = pickle.dumps(f)
f2 = pickle.loads(b)

assert(f2['Five'] == 5)
assert(f2.type == f.type)

sys.exit(0)




from . import indexmod, pipesegment, G3FrameType, log_fatal, G3Reader, G3NetworkSender

@indexmod
def Delete(frame, keys=[], type=None):
    '''
    Deletes specified keys from frame. If optional type specified, only acts on frames of the given type.
    '''
    if type is not None and frame.type != type:
        return

    for key in keys:
        if key in frame:
            del frame[key]

@indexmod
def Rename(frame, keys={}, type=None):
    '''
    Renames specified keys in frame. If optional type specified, only acts on frames of the given type. Argument is a dictionary mapping old names to new ones.
    '''
    if type is not None and frame.type != type:
        return

    for key in keys:
        if key in frame:
            frame[keys[key]] = frame[key]
            del frame[key]

@indexmod
def Dump(frame, type=None, added_message = None):
    '''
    Prints frames to console. If optional type specified, only acts on frames of the given type.
    '''
    if type is not None and frame.type != type:
        return
    if frame.type == G3FrameType.EndProcessing:
        return
    if added_message:
        print(added_message)
    print(frame)


@indexmod
class InjectFrame(object):
    """
    Inject an arbitrary frame into a pipeline.

    Arguments
    ---------
    frame : G3Frame
        The frame to inject
    """
    def __init__(self, frame):
        self.frame = frame
    def __call__(self, frame):
        if self.frame is None:
            return
        out = [self.frame, frame]
        self.frame = None
        return out


@indexmod
def InjectDebug(frame, type=None, debug_start_func = None):
    '''Starts a pdb session when a frame of type shows up.

    The frame data is stored in the variable named "frame".

    If ``debug_start_func`` is not None, only starts a debug session when
    ``debug_start_func(frame) == True``.
    '''
    if type is None or frame.type == type:
        if ((debug_start_func is None) or debug_start_func(frame)):
            import pdb, rlcompleter
            pdb.Pdb.complete = rlcompleter.Completer(locals()).complete
            pdb.set_trace()


@indexmod
class AbortAfterNFrames(object):
    '''Stops processing after n_frames frames go by'''
    def __init__(self, type, n_frames):
        self.n_desired_frames = n_frames
        self.num_frames = 0
        self.type = type
    def __call__(self, frame):
        if self.num_frames >= self.n_desired_frames:
            log_fatal("Manual Abort Triggered")
        if frame.type == self.type:
            self.num_frames += 1
        return

@pipesegment
def G3NetworkReceiver(pipe, hostname='localhost', port=5978):
    '''
    Emulation of old G3NetworkReceiver class. Equivalent to pointing
    G3Reader at a TCP URL.
    '''

    pipe.Add(G3Reader, filename='tcp://' + hostname + ':' + str(port))

@indexmod
class G3ThrottledNetworkSender(object):
    '''
    Send every Nth frame of certain types using a wrapped G3NetworkSender.
    All instances of frames not in the dictionary frame_decimation will be sent
    at their full rate.
    '''
    def __init__(self, hostname='*', port=5978, frame_decimation = {G3FrameType.Timepoint: 10}, max_queue_size=0):
        self.sender = G3NetworkSender(hostname=hostname, port=port, max_queue_size=max_queue_size)
        self.decimation = frame_decimation
        self.counts = {}
        for k in self.decimation.keys():
            self.counts[k] = 0
    def __call__(self, frame):
        if frame.type in self.counts:
            if self.decimation[frame.type] == 0:
                return
            self.counts[frame.type] += 1
            if self.counts[frame.type] % self.decimation[frame.type] != 0:
                return

        self.sender(frame)

@indexmod
class DeduplicateMetadata(object):
    '''
    Drop metadata frames (e.g. Calibration, Wiring) for which the previous
    metadata frame of the same type is byte-for-byte identical. This can be
    handy when, for example, reading in many files from the G3MultiFileWriter,
    which copies metadata frames to the beginning of each file. Considers
    all frames not in <dataframetypes> to be metadata (by default, everything
    except Timepoint and Scan frames).
    '''
    def __init__(self, dataframetype=[G3FrameType.Timepoint, G3FrameType.Scan]):
        self.dataframes = dataframetype
        self.metacache = {}
    def __call__(self, f):
        if f.type in self.dataframes or f.type == G3FrameType.EndProcessing:
            # Pass data or end processing frames through
            return

        # Compare serialized form of frame
        from hashlib import md5
        rawdata = md5(f.hash).hexdigest()
        if f.type in self.metacache and self.metacache[f.type] == rawdata:
            # Same as the last one, so drop this frame
            return []

        # Otherwise, pass through and cache
        self.metacache[f.type] = rawdata

@indexmod
class DropOrphanMetadata(object):
    '''
    Remove metadata frames (e.g. Calibration, Wiring) without
    intervening data frames (e.g. Timepoint, Scan, specified by the
    <dataframetype> argument to the constructor). The metadata frames that
    do show up will be the most recent of each type and appear in their
    original order.
    '''
    def __init__(self, dataframetype=[G3FrameType.Timepoint, G3FrameType.Scan]):
        self.dataframes = dataframetype
        self.metacache = []
    def __call__(self, frame):
        if frame.type == G3FrameType.EndProcessing:
            return
        if frame.type in self.dataframes:
            # If we got a data frame, dump any queued metadata before it
            # and clear the queue.
            out = self.metacache + [frame]
            self.metacache = []
            return out
        else:
            # Insert new metadata frames into the same spot in the queue
            # as the last instance of a frame of that type or the end of the
            # queue if there is no such frame.
            for i in range(len(self.metacache)):
                if self.metacache[i].type == frame.type:
                    self.metacache[i] = frame
                    break
            else:
                self.metacache.append(frame)
            return []
                    

del indexmod
del pipesegment



#!/usr/bin/env python
from spt3g import core
import sys, glob

class NumberAdder(object):
    def __init__(self):
        self.n = 0
    def __call__(self, frame):
        if frame.type == core.G3FrameType.EndProcessing:
            return
        frame['TheNumber'] = self.n
        self.n += 1

#set up our file to read
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
pipe.Add(core.G3Writer, filename = 'testlazyreader.g3')
pipe.Run()

na = NumberAdder()
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename = 'testlazyreader.g3', n_frames_to_read = 3)
pipe.Add(na)
pipe.Run() 

if na.n != 3:
    print("N Reader failed", na.n)
    sys.exit(1)





#!/usr/bin/env python

import pickle, numpy, sys
from spt3g import core

dat = numpy.floor(numpy.random.normal(size=1200, scale=20, loc=0))
ts = core.G3Timestream(dat)
ts.units = core.G3TimestreamUnits.Counts
ts.SetFLACCompression(5)
ts.SetFLACBitDepth(24)
ts[5] = numpy.nan
dat[5] = numpy.nan

pickle.dump(ts, open('tsdump.pkl', 'wb'))

ts_rehyd = pickle.load(open('tsdump.pkl', 'rb'))
print('Original')
print(ts.units)
print(ts[4], ts[5], ts[6])
print('Rehydrated')
print(ts_rehyd.units)
print(ts_rehyd[4], ts_rehyd[5], ts_rehyd[6])

if ts_rehyd.units != ts.units:
	print('Units do not match')
	sys.exit(1)

if numpy.isfinite(ts_rehyd[5]):
	print('Element 5 finite!')
	sys.exit(1)

if (numpy.asarray(ts_rehyd)[numpy.isfinite(ts_rehyd)] != dat[numpy.isfinite(dat)]).any():
	print('Elements not preserved!')
	sys.exit(1)




import numpy
from . import G3Timestream, DoubleVector, G3VectorDouble, G3TimestreamMap, G3VectorTime, G3Time, Int64Vector, G3VectorInt, \
    G3VectorComplexDouble, ComplexDoubleVector, BoolVector, G3VectorBool, IntVector, FloatVector, ComplexFloatVector
from . import G3Units, log_fatal, log_warn, usefulfunc, G3FrameObject

__all__ = ['concatenate_timestreams']

# Use numpy bindings rather than vector_indexing_suite for G3Timestream
def G3Timestream_getitem(x, y):
    if isinstance(y, slice):
        # If the argument is a slice, and the value would be a numpy array,
        # turn it into an appropriately sliced timestream, with appropriate
        # start and stop times. This is done in C++ for speed.
        it = x._cxxslice(y)
    else:
        it = numpy.asarray(x).__getitem__(y)

    return it

@property
def timestream_dtype(ts):
    """
    Numpy data type of stored data.
    """
    return numpy.asarray(ts).dtype

G3Timestream.__getitem__ = G3Timestream_getitem
G3Timestream.__setitem__ = lambda x, y, z: numpy.asarray(x).__setitem__(y, z)
G3Timestream.dtype = timestream_dtype
G3TimestreamMap.dtype = timestream_dtype

def timestreamastype(a, dtype):
    '''
    Convert timestream to a different data type. See numpy.array.astype()
    '''
    if a.dtype == numpy.dtype(dtype):
        return a
    out = G3Timestream(numpy.asarray(a).astype(dtype))
    out.units = a.units
    out.start = a.start
    out.stop = a.stop
    return out
G3Timestream.astype = timestreamastype

def timestreammapastype(a, dtype):
    '''
    Convert timestream map to a different data type.  See numpy.array.astype()
    '''
    if a.dtype == numpy.dtype(dtype):
        return a
    return G3TimestreamMap(a.keys(), numpy.asarray(a).astype(dtype), a.start, a.stop, a.units, 0, False)
G3TimestreamMap.astype = timestreammapastype

# XXX consider replacing all of this with the numpy __array_ufunc__ machinery

# Provide all the numpy binary arithmetic operators, first non-in-place, then
# in-place, with slightly different semantics
def numpybinarywrap(a, b, op):
    is_g3 = isinstance(a, G3FrameObject) or isinstance(b, G3FrameObject)
    is_tsa = isinstance(a, G3Timestream)
    is_tsb = isinstance(b, G3Timestream)
    is_cxa = isinstance(a, (G3VectorComplexDouble, ComplexDoubleVector, ComplexFloatVector))
    is_cxb = isinstance(b, (G3VectorComplexDouble, ComplexDoubleVector, ComplexFloatVector))
    cls = a.__class__
    if is_tsa:
        if is_tsb:
            a._assert_congruence(b)
        elif is_cxb:
            return NotImplemented
    elif is_cxa:
        pass
    elif is_tsb or is_cxb:
        return NotImplemented
    out = op(numpy.asarray(a), numpy.asarray(b))
    k = out.dtype.kind
    s = out.dtype.itemsize
    if k == 'c':
        if is_g3:
            return G3VectorComplexDouble(out)
        return ComplexDoubleVector(out) if s == 16 else ComplexFloatVector(out)
    elif k == 'b':
        return G3VectorBool(out) if is_g3 else BoolVector(G3VectorBool(out))
    elif is_tsa:
        out = G3Timestream(out)
        out.units = a.units
        out.start = a.start
        out.stop = a.stop
        return out
    elif k == 'f':
        if is_g3:
            return G3VectorDouble(out)
        return DoubleVector(out) if s == 8 else FloatVector(out)
    elif k in 'iu':
        out = out.astype(int)
        if is_g3:
            return G3VectorInt(out)
        return Int64Vector(out) if s == 8 else IntVector(out)
    return NotImplemented

def numpyinplacebinarywrap(a, b, op):
    if isinstance(a, G3Timestream) and isinstance(b, G3Timestream):
        a._assert_congruence(b)
    op(numpy.asarray(a), numpy.asarray(b))
    return a

all_cls = [G3Timestream, G3VectorDouble, DoubleVector, G3VectorInt, Int64Vector,
           G3VectorComplexDouble, ComplexDoubleVector, G3VectorBool, BoolVector,
           FloatVector, IntVector, ComplexFloatVector]

for attr in ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow',
             'and', 'or', 'xor', 'lshift', 'rshift']:
    for cls in all_cls:
        x = '__{}__'.format(attr)
        if x in numpy.ndarray.__dict__:
            setattr(cls, x,
                    lambda a, b, op=numpy.ndarray.__dict__[x]: numpybinarywrap(a, b, op))
        x = '__r{}__'.format(attr)
        if x in numpy.ndarray.__dict__:
            setattr(cls, x,
                    lambda a, b, op=numpy.ndarray.__dict__[x]: numpybinarywrap(a, b, op))
        x = '__i{}__'.format(attr)
        if x in numpy.ndarray.__dict__:
            setattr(cls, x,
                    lambda a, b, op=numpy.ndarray.__dict__[x]: numpyinplacebinarywrap(a, b, op))

# unary operators
for x in ['__neg__', '__pos__', '__invert__', '__abs__', '__bool__']:
    if x not in numpy.ndarray.__dict__:
        continue
    for cls in all_cls:
        setattr(cls, x,
                lambda a, op=numpy.ndarray.__dict__[x]: a.__class__(op(numpy.asarray(a))))

# Bind some useful nativish binary operators
for x in ['__eq__', '__ge__', '__gt__', '__le__', '__lt__', '__ne__']:
    if x not in numpy.ndarray.__dict__:
        continue
    for cls in all_cls:
        setattr(cls, x,
                lambda a, b, op=numpy.ndarray.__dict__[x]: numpybinarywrap(a, b, op))

# Bind some useful methods
for x in ["sum", "mean", "any", "all", "min", "max", "argmin", "argmax", "var", "std"]:
    if x not in numpy.ndarray.__dict__:
        continue
    for cls in all_cls:
        def ufunc_wrapper(op):
            def ufunc(a, *args, **kwargs):
                return numpy.ndarray.__dict__[op](numpy.asarray(a), *args, **kwargs)
            ufunc.__doc__ = numpy.ndarray.__dict__[op].__doc__
            return ufunc
        setattr(cls, x, ufunc_wrapper(x))

# Bind some memory-efficient methods for G3TimestreamMap
for op in ["std", "var"]:
    def ufunc_wrapper(op):
        def ufunc(a, axis=None, dtype=None, out=None, **kwargs):
            bound_args = {}
            if op in ["std", "var"]:
                bound_args["ddof"] = kwargs.pop("ddof", 0)
            if not len(kwargs) and (axis == -1 or axis == 1):
                ret = numpy.asarray(getattr(a, "_c" + op)(**bound_args))
                if out is not None:
                    out[:] = ret
                    return out
                if dtype is not None:
                    return ret.astype(dtype)
                return ret
            kwargs.update(**bound_args)
            return numpy.ndarray.__dict__[op](numpy.asarray(a), axis, dtype, out, **kwargs)
        ufunc.__doc__ = numpy.ndarray.__dict__[op].__doc__
        return ufunc

    setattr(G3TimestreamMap, op, ufunc_wrapper(op))


#add concatenation routines to g3timestream objects
def _concatenate_timestreams(cls, ts_lst, ts_rounding_error=0.6, ts_interp_threshold=0):
    """
    Concatenate G3Timestream objects together.

    Arguments
    ---------
    ts_lst : list
        list of G3Timestream objects.
    ts_rounding_error : float
        allowed error in timestream separation such that timestreams are
        contiguous, as a fraction of the sample rate. This should be
        0 by default, but is 0.5 to allow for downsampler shifting,
        and then bumpted again to 0.6 to allow for floating-point
        errors in what 0.5 is.
    ts_interp_threshold : float
        allowed timestream separation below which gaps between timestreams are
        interpolated to be made continuous

    Returns
    -------
    ts : G3Timestream instance
        The concatenation of the input list of G3Timestream objects
    """
    #check for contiguous timestreams
    for i in range(1, len(ts_lst)):
        ts_sep = (ts_lst[i].start.time - ts_lst[i-1].stop.time) * ts_lst[i].sample_rate
        if numpy.abs(ts_sep - 1) > ts_rounding_error:
            if (ts_interp_threshold > 0) and ((ts_sep - 1) < ts_interp_threshold):
                log_warn("Timestreams are not contiguous: timestreams %d and %d "
                          "separated by %f samples (%s).  Interpolating." %
                         (i, i-1, ts_sep - 1, str(ts_lst[i].start)))
                v = numpy.linspace(ts_lst[i-1][-1], ts_lst[i][0], ts_sep + 1)[1:-1]
                ts_interp = cls(v)
                ts_interp.units = ts_lst[0].units
                ts_interp.start = ts_lst[i-1].stop + 1. / ts_lst[i].sample_rate
                ts_interp.stop = ts_lst[i].start - 1. / ts_lst[i].sample_rate
                ts_lst = ts_lst[:i] + [ts_interp] + ts_lst[i:]
            else:
                log_fatal("Timestreams are not contiguous: timestreams %d and %d "
                          "separated by %f samples (%s)" % (i, i-1, ts_sep - 1, str(ts_lst[i].start)))
        if (ts_lst[i].units != ts_lst[0].units):
            log_fatal("Timestreams are not the same units")
    out_ts = cls(numpy.concatenate(ts_lst))
    out_ts.units = ts_lst[0].units
    out_ts.start = ts_lst[0].start
    out_ts.stop = ts_lst[-1].stop
    return out_ts

G3Timestream.concatenate = classmethod(_concatenate_timestreams)


def _concatenate_timestream_maps(cls, ts_map_lst, ts_rounding_error=0.6, ts_interp_threshold=0, skip_missing=False):
    """
    Concatenate G3TimestreamMap objects together.

    Arguments
    ---------
    ts_map_lst : list
        list of G3TimestreamMap objects.
    ts_rounding_error : float
        allowed error in timestream separation such that timestreams are
        contiguous, as a fraction of the sample rate. This should be
        0 by default, but is 0.5 to allow for downsampler shifting,
        and then bumpted again to 0.6 to allow for floating-point
        errors in what 0.5 is.
    ts_interp_threshold : float
        allowed timestream separation below which gaps between timestreams are
        interpolated to be made continuous
    skip_missing : bool
        If True, include only the channels that are present in all of the
        input G3TimestreamMap object.  Otherwise, raises an error if any
        map does not have the same keys.

    Returns
    -------
    tsm : G3TimestreamMap instance
        The concatenation of the input list of G3TimestreamMap objects
    """
    keys = list(ts_map_lst[0].keys())
    skeys = set(keys)
    for tsm in ts_map_lst[1:]:
        if skip_missing:
            skeys &= set(tsm.keys())
            continue
        if set(tsm.keys()) != skeys:
            log_fatal("Timestream maps do not have the same keys")
    if skip_missing:
        # preserve key order
        keys = [k for k in keys if k in skeys]
    out_tsm = cls()
    for k in keys:
        ts_lst = [tsm[k] for tsm in ts_map_lst]
        out_tsm[k] = G3Timestream.concatenate(ts_lst, ts_rounding_error, ts_interp_threshold)
    return out_tsm

G3TimestreamMap.concatenate = classmethod(_concatenate_timestream_maps)


# global concatenation function
@usefulfunc
def concatenate_timestreams(ts_lst, ts_rounding_error=0.6, ts_interp_threshold=0):
    """
    Concatenate G3Timestream or G3TimestreamMap objects together.

    Arguments
    ---------
    ts_lst : list
        list of G3Timestream or G3TimestreamMap objects.  Must all be
        the same type.
    ts_rounding_error : float
        allowed error in timestream separation such that timestreams are
        contiguous, as a fraction of the sample rate. This should be
        0 by default, but is 0.5 to allow for downsampler shifting,
        and then bumpted again to 0.6 to allow for floating-point
        errors in what 0.5 is.
    ts_interp_threshold : float
        allowed timestream separation below which gaps between timestreams are
        interpolated to be made continuous

    Returns
    -------
    ts : G3Timestream or G3TimestreamMap object
        The concatenation of the input list of objects
    """
    return ts_lst[0].concatenate(ts_lst, ts_rounding_error, ts_interp_threshold)


@property
def tsm_names(self):
    '''
    Get timestream map channel names.
    '''
    return list(self.keys())

G3TimestreamMap.names = tsm_names

@property
def tsm_data(self):
    '''
    Return a numpy array view into the underlying 2D array of the timestream map
    '''
    return numpy.asarray(self)

G3TimestreamMap.data = tsm_data



from .._libdfmux import *

from .HardwareMapTools import (
    GenerateFakeHardwareMap,
    PyDfMuxHardwareMapInjector,
    PyDfMuxHardwareMapInjectorAllChannels,
    DfmlHardwareMapInjector,
    PyDfMuxBolometerPropertiesInjector,
    PyDfMuxWiringMapInjector,
    PyDfMuxWiringMapInjectorAllChannels,
    PathStringForBolo,
)
from .ScanTools import FixedLengthScans
from .Housekeeping import HousekeepingConsumer, PeriodicHousekeepingCollector, HousekeepingForBolo
from .LegacyHousekeeping import LegacyHousekeepingConsumer
from .unittransforms import get_timestream_unit_conversion, ConvertTimestreamUnits
from .DataQualityTools import FillMissingTimepointFrames



import socket, json

#
# Implement just enough asynchronous HTTP to let us issue Tuber calls.
# This does not use urllib or something because these libraries do not let
# you issue a request and get the reply later.
#

class TuberClient(object):
    def __init__(self, hostname, timeout=None):
        self.host = hostname
        self.socket = None
        self.timeout = timeout
    def OpenSocket(self):
        self.socket = socket.socket()
        self.socket.connect((self.host, 80))
        self.socket.settimeout(self.timeout)
    def ReadLineFromSocket(self):
        out = str()
        while not out.endswith('\n'):
            out += self.socket.recv(1).decode()
        return out
    def CallMethod(self, obj, method, wait_for_reply=True):
        if self.socket is None:
            self.OpenSocket()

        request = 'POST /tuber HTTP/1.1\r\n'
        request += 'Connection: keep-alive\r\n'
        request += 'Host: %s\r\n' % self.host
        request += 'Content-Type: application/json\r\n'

        if not isinstance(method, str) and hasattr(method, '__iter__'):
            jsonbits = '[%s]' % ', '.join(['{"object": "%s", "method": "%s"}' % (obj, m) for m in method])
        else:
            jsonbits = '[{"object": "%s", "method": "%s"}]' % (obj, method)
        request += 'Content-Length: %d\r\n\r\n' % len(jsonbits)
        request += jsonbits

        self.socket.send(request.encode())

        if wait_for_reply:
            return self.GetReply()
    def GetReply(self):
        toread = None
        chunked = False
        connection_atend = 'close'
        while True:
            header = self.ReadLineFromSocket().strip()
            if header == '':
                break
            header = header.split(':')
            if header[0] == 'Content-Length':
                toread = int(header[1].strip())
            if header[0] == 'Transfer-Encoding' and header[1].strip() == 'chunked':
                chunked = True
            if header[0] == 'Connection':
                connection_atend = header[1].strip()
        assert(toread is not None or chunked)
        assert(toread is None or chunked is False)
        out = ''

        if chunked:
            still_chunking = True
            while still_chunking:
                chunksize = self.ReadLineFromSocket().strip()
                toread = int(chunksize.split(';')[0], base=16)
                if toread == 0:
                    still_chunking = False
                while toread > 0:
                    chunk = self.socket.recv(toread).decode()
                    assert(len(chunk) > 0)
                    out += chunk
                    toread -= len(chunk)
                lf = self.ReadLineFromSocket().strip()
                assert(len(lf)) == 0
        else:
            while toread > 0:
                chunk = self.socket.recv(toread).decode()
                assert(len(chunk) > 0)
                out += chunk
                toread -= len(chunk)
        if connection_atend == 'close':
            self.socket.close()
            self.socket = None
        return json.loads(out)




#!/usr/bin/env python

from spt3g import core, dfmux, gcp
import socket, argparse, os

parser = argparse.ArgumentParser(description='''
Record dfmux data to disk.

Two modes of operation are suppored.

The standard mode is to supply a hardware map from which to determine which
IceBoards to communicate with.

An alternative mode is to also supply a list of IceBoard serial numbers.  In
that case, the HWM is not loaded from disk (saving as much as several minutes,
depending on the size of the HWM), and the serial numbers are used directly to
initialize the DAQ.  The hardware map file path is instead used to find a
nominal_online_cal.g3 file in the same directory as the YAML file, which is
injected into the DAQ stream for teeing to live visualization tools like
Lyrebird.

Optional auxiliary modules such as the SPT calibrator or the Polarbear WHWP
encoder stream are not enabled by default.

Two methods of error monitoring are also available: enabling the syslog logger
sends all output log messages to a central rsyslog database (e.g. as set up by
SPT), and enabling the GCP watchdog module pings the SPT pager server
periodically to indicate normal operation.
'''
)
parser.add_argument('hardware_map', metavar='/path/to/hwm.yaml', help='Path to hardware map YAML file')
parser.add_argument('boards', nargs='*', metavar='serial', help='IceBoard serial(s) from which to collect data')
parser.add_argument('output', metavar='/path/to/files/', help='Directory in which to place output files')

parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode (print all frames)')
parser.add_argument('-u', '--udp', action='store_true', help='Use UDP multicast instead of SCTP')
parser.add_argument('--daq-threads', default=None, type=int, help='Number of listener threads to use (applies only for SCTP)')
parser.add_argument('--max-file-size', default=1024, type=int, help='Maximum file size in MB (default 1024)')
parser.add_argument('--calibrator', action='store_true', help='Enable SPT calibrator DAQ')
parser.add_argument('--whwp', action='store_true', help='Enable Polarbear HWP DAQ')
parser.add_argument('--syslog', action='store_true', help='Enable syslog logger')
parser.add_argument('--watchdog', action='store_true', help='Enable GCP watchdog ping')
args = parser.parse_args()

# Tee log messages to both log file and GCP socket
console_logger = core.G3PrintfLogger()
console_logger.timestamps = True # Make sure to get timestamps in the logs

if args.syslog:
    import syslog
    core.G3Logger.global_logger = core.G3MultiLogger(
        [console_logger, core.G3SyslogLogger("dfmuxdaq: ", syslog.LOG_USER)]
    )
else:
    core.G3Logger.global_logger = console_logger

core.set_log_level(core.G3LogLevel.LOG_ERROR, 'DfMuxBuilder')
core.set_log_level(core.G3LogLevel.LOG_ERROR, 'DfMuxCollector')

args.hardware_map = os.path.realpath(args.hardware_map)

if not len(args.boards):
    # If the input is a hardware map path, import the HWM and
    # extract the list of boards from it

    core.log_notice('Initializing hardware map and boards',
                    unit='Data Acquisition')
    import pydfmux
    hwm = pydfmux.load_session(open(args.hardware_map, 'r'))['hardware_map']
    boards = hwm.query(pydfmux.Dfmux)
    boards.resolve()
    boards = boards.serial

else:
    # Otherwise assume the input is a list of board serials
    core.log_notice('Acquiring hardware map information from boards',
                    unit='Data Acquisition')
    hwm = None
    boards = ['%04d' % (int(b)) for b in args.boards]

core.log_notice('Beginning data acquisition', unit='Data Acquisition')
# Set up DfMux consumer
pipe = core.G3Pipeline()
builder = dfmux.DfMuxBuilder([int(board) for board in boards])

if args.udp:
    # Set up listeners per thread and point them at the event builder
    # Get the local IP(s) to use to connect to the boards by opening test
    # connections. Using a set rather than a list deduplicates the results.
    local_ips = {}
    for board in boards:
        testsock = socket.create_connection(('iceboard' + board + '.local', 80))
        local_ip = testsock.getsockname()[0]
        if local_ip not in local_ips:
            local_ips[local_ip] = set()
        local_ips[local_ip].add(int(board))
        testsock.close()
    core.log_notice('Creating listeners for %d boards on interfaces: %s' %
                    (len(boards), ', '.join(local_ips.keys())),
                    unit='Data Acquisition')

    # Set up listeners per network segment and point them at the event builder
    collectors = [dfmux.DfMuxCollector(ip, builder, list(local_boards))
                  for ip, local_boards in local_ips.items()]
else:
    iceboard_hosts = ['iceboard' + board + '.local' for board in boards]
    if args.daq_threads is None:
        args.daq_threads = len(iceboard_hosts)
    if args.daq_threads > len(iceboard_hosts):
        args.daq_threads = len(iceboard_hosts)
    collectors = [dfmux.DfMuxCollector(builder, iceboard_hosts[i::args.daq_threads])
                  for i in range(args.daq_threads)]
pipe.Add(builder)

# Catch errors if samples have become misaligned. Don't even bother processing
# data involving a significant (N-2) reduction from the boards we should have.
nboards = len(boards)
n_badpkts = 0
n_goodpkts = 0
logtweaked = False
def yellanddrop(fr):
    global n_badpkts
    global n_goodpkts
    global logtweaked

    if fr.type != core.G3FrameType.Timepoint:
        return

    if len(fr['DfMux']) < nboards-2:
        if n_badpkts > 0 and n_badpkts % 100 == 0:
            core.log_error('Only %d/%d boards (%s) responding for %d samples -- check for sample misalignment. Temporarily suppressing DfMuxBuilder logging and disabling data archiving.' % (len(fr['DfMux']), nboards, ', '.join([str(k) for k in fr['DfMux'].keys()]), n_badpkts), unit='Data Acquisition')
            # Turn up the threshold on DfMuxBuilder to prevent flooding the console
            core.set_log_level(core.G3LogLevel.LOG_ERROR, 'DfMuxBuilder')
            logtweaked = True
        n_badpkts += 1
        n_goodpkts = 0
        return []
    else:
        n_goodpkts += 1
        if n_goodpkts > 5 and logtweaked:
            # Turn the threshold back down
            core.set_log_level(core.G3LogLevel.LOG_NOTICE, 'DfMuxBuilder')
            core.log_notice('Gross board misalignment resolved. Re-enabling DfMuxBuilder logging and data archiving.', unit='Data Acquisition')
            logtweaked = False
            n_badpkts = 0
pipe.Add(yellanddrop)

# Collect housekeeping when GCP asks for it and send it back to GCP when desired
pipe.Add(gcp.GCPSignalledHousekeeping)
pipe.Add(dfmux.HousekeepingConsumer)
pipe.Add(gcp.GCPHousekeepingTee)

def WaitForWiring(frame):
    if frame.type == core.G3FrameType.Wiring:
        core.set_log_level(core.G3LogLevel.LOG_NOTICE, 'DfMuxBuilder')
        core.set_log_level(core.G3LogLevel.LOG_NOTICE, 'DfMuxCollector')
        core.log_notice('Got a wiring frame, ready for data acquisition.', unit='Data Acquisition')
pipe.Add(WaitForWiring)

# For visualization, add nominal pointing
if hwm is None:
    # Get the bolometer properties map from disk (written separately by pydfmux)
    # whenever a new wiring frame is received.
    def BolometerPropertiesInjector(frame):
        if frame.type == core.G3FrameType.Wiring:
            nchan = len(frame['WiringMap'].keys())
            core.log_notice("Collecting data from %d mapped channels." % (nchan),
                            unit='Data Acquisition')
            bpm = os.path.join(os.path.dirname(args.hardware_map), 'nominal_online_cal.g3')
            try:
                if not os.path.exists(bpm):
                    raise IOError('Missing file %s' % bpm)
                fr = list(core.G3File(bpm))[0]
                return [frame, fr]
            except Exception as e:
                core.log_warn('Error loading BolometerPropertiesMap: %s' % (str(e)),
                              unit='Data Acquisition')
    pipe.Add(BolometerPropertiesInjector)
else:
    # Otherwise inject using the loaded HWM
    def CheckWiring(frame):
        if frame.type == core.G3FrameType.Wiring:
            nchan = len(frame['WiringMap'].keys())
            core.log_notice("Collecting data from %d mapped channels." % (nchan),
                            unit='Data Acquisition')
    pipe.Add(CheckWiring)
    pipe.Add(dfmux.PyDfMuxBolometerPropertiesInjector, pydfmux_hwm=hwm)

# Provide a tee for realtime visualization before possible buffering
# in the calibrator DAQ code
pipe.Add(core.G3ThrottledNetworkSender, hostname='*', port=5451, max_queue_size=1000)

# Collect data from the SPT chopped calibration source
if args.calibrator:
    from spt3g import auxdaq
    pipe.Add(auxdaq.CalibratorDAQ)

# Collect data from the Polarbear warm half wave plate angular encoder
if args.whwp:
    from spt3g import whwp
    pipe.Add(whwp.WHWPConsumer)

# Issue a periodic watchdog ping to the SPT pager system
if args.watchdog:
    pipe.Add(gcp.DAQWatchdog, calibrator=args.calibrator)

if args.verbose:
    pipe.Add(core.Dump)

# Provide a tee for other software to collect frames
pipe.Add(core.G3NetworkSender, hostname='*', port=5353, max_queue_size=1000)

def filename(frame, seq):
    if 'EventHeader' in frame:
        t = frame['EventHeader']
    else:
        t = core.G3Time.Now()
    return os.path.join(args.output, t.GetFileFormatString() + '.g3')
pipe.Add(core.G3MultiFileWriter, filename=filename, size_limit=args.max_file_size*1024*1024)

for collector in collectors:
    collector.Start()
pipe.Run(profile=True)

# Shut everything down
for collector in collectors:
    collector.Stop()

# C++ global destructor runs after python has closed, sometimes
# Setting the logger to None like this explicitly avoids segfaults
core.G3Logger.global_logger = None



from ..core import G3Units, Quat, G3VectorQuat, G3TimestreamQuat, usefulfunc, indexmod
import numpy


@usefulfunc
def quat_to_ang(q):
    """
    Convert a pointing quaternion (or vector of them) to a set of angles (or
    vector of them) specified as a (longitude, latitude) pair.
    """
    single = False
    if isinstance(q, Quat):
        q = numpy.array(G3VectorQuat([q]))
        single = True
    elif isinstance(q, list):
        q = numpy.array(G3VectorQuat(q))
    else:
        q = numpy.array(q)

    # Copied from C code
    d = q[:, 1] ** 2 + q[:, 2] ** 2 + q[:, 3] ** 2
    mask = numpy.abs(d - 1.0) > 1e-6
    q[mask] /= d[mask][:, None] ** 0.5

    delta = numpy.arcsin(q[:, 3]) * G3Units.rad
    alpha = numpy.arctan2(q[:, 2], q[:, 1]) * G3Units.rad

    if single:
        return (alpha[0], delta[0])
    else:
        return (alpha, delta)


@usefulfunc
def ang_to_quat(alpha, delta, start=None, stop=None):
    """
    Convert a set of angles (or vector of them) specified as a (longitude,
    latitude) pair to a pointing quaternion (or vector of them). If start
    and stop are defined, the return value for vectors will be a
    G3TimestreamQuat with start and stop times set to the provided values.
    """

    alpha = numpy.asarray(alpha) / G3Units.rad
    delta = numpy.asarray(delta) / G3Units.rad
    # Copied from C code
    c_delta = numpy.cos(delta)
    q = numpy.column_stack(
        (
            0 * c_delta,  # 0s with the right shape
            c_delta * numpy.cos(alpha),
            c_delta * numpy.sin(alpha),
            numpy.sin(delta),
        )
    )

    if len(q) == 1:
        return Quat(q[0, :])
    else:
        if start is not None:
            out = G3TimestreamQuat(q)
            out.start = start
            out.stop = stop
            return out
        else:
            return G3VectorQuat(q)

@indexmod
def AddTimingToPointingQuats(fr, key, timing_ref='RawBoresightAz'):
    """
    Transforms a quaternion pointing timestream <key> from a G3VectorQuat
    to a G3TimestreamQuat by adding timing information extracted from some
    other timestream-like object (scalar pointing timestreams, detector
    timestreams etc.) specified by <timing_ref>. Because this operation is
    backwards-compatible and involves no data loss, the transformation is
    done in-place -- the previous data are deleted and replaced with the new
    one.
    """
    
    if key not in fr:
        return

    if isinstance(fr[key], G3TimestreamQuat):
        return

    x = fr.pop(key)
    x = G3TimestreamQuat(x)
    x.start = fr[timing_ref].start
    x.stop = fr[timing_ref].stop
    fr[key] = x




import json

def uniquify_list(seq):  
    checked = [] 
    for e in seq: 
        if e not in checked: 
            checked.append(e) 
    return checked

def frame_type_str_to_color(s):
    cmap = {'Timepoint': 'm',
            'Housekeeping' : 'y',
            'Scan' : 'r',
            'Map' : 'g',
            'InstrumentStatus' : 'k',
            'Wiring' : 'c',
            'Calibration' : 'b',
            'EndProcessing' : 'k',
            'None' : 'k',
        }
    assert(s in cmap)
    return cmap[s]

def plot_frame_processing_info(g3_pipeline):
    import matplotlib.pyplot as plt

    plt.clf()
    parsed_output = json.loads( g3_pipeline.GetGraphInfo()  )

    proc_lst_trans = zip(*parsed_output['Processing List'])
    ptime = proc_lst_trans[0]
    pmod = proc_lst_trans[1]
    pframe = proc_lst_trans[2]
    ptype = proc_lst_trans[3]

    n_frames = max(pframe)+1
    n_times = max(ptime)+1
    n_mods = max(pmod)+1


    #create the legend in a realllly hacky way
    unique_ptypes = uniquify_list(ptype)
    ax = plt.gca()
    colors = map(frame_type_str_to_color, unique_ptypes)
    labels = unique_ptypes
    [plt.plot([], [], color=c,label=l, linewidth=3 )[0] for c, l in zip(colors,labels)]
    plt.legend(labels)

    #these store the information for the state of the frame
    frame_mod_pos = {}
    frame_prev_time = {}

    #plots the data flow
    for i in range(len(ptime)):
        if pframe[i] not in frame_mod_pos:
            frame_mod_pos[pframe[i]] = pmod[i] - 1
        prev_mod = frame_mod_pos[pframe[i]]
        if pframe[i] in frame_prev_time:
            prev_time = frame_prev_time[pframe[i]]
        else:
            prev_time = ptime[i]
        cur_mod = pmod[i]
        cur_time = ptime[i]
        frame_id = pframe[i]
        x_pos = lambda mod: mod - float(frame_id + 1)/(3*float(n_frames + 1)) + .66
        plt.plot([x_pos(prev_mod), x_pos(prev_mod)], 
                 [prev_time, cur_time], 
                 color = frame_type_str_to_color(ptype[i]),
                 linestyle = '--'
        )
        plt.plot([x_pos(prev_mod), x_pos(cur_mod)], 
                 [cur_time, cur_time+1],
                 color = frame_type_str_to_color(ptype[i]),
                 linewidth = 2.0,
                 marker = '.'
             )
        frame_mod_pos[pframe[i]] = cur_mod
        frame_prev_time[pframe[i]] = cur_time + 1

    #handles setting up labels
    mod_x_ticks = []
    mod_x_labels = []

    module_stripper = lambda mods_name: mods_name[mods_name.rfind('.')+1:]
    for p in parsed_output["Module List"]:
        plt.axvline(p[0], linewidth = 2, linestyle = '-.', color = 'k')
        mod_x_ticks.append(p[0]+.5)
        mod_x_labels.append(module_stripper(p[1]))

    plt.xticks(mod_x_ticks, mod_x_labels, rotation=-7)    
    #does some silly formatting
    plt.gca().get_yaxis().set_ticks([])
    plt.ylim(n_times+.5, -0.5)
    plt.xlim(0, n_mods)
    plt.ylabel("Time\n$\\Longleftarrow$")
    plt.xlabel("Module $\\Longrightarrow$")
    plt.show()




from spt3g import core
import time

pipe = core.G3Pipeline()

pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
def addtime(fr):
	fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
def printfr(fr):
	if fr.type != core.G3FrameType.EndProcessing:
		print(fr)
pipe.Add(addtime)
pipe.Add(printfr) # Could use core.Dump as well
pipe.Add(core.G3Writer, filename='/tmp/test2.g3')

pipe.Run()



#!/usr/bin/env python
from __future__ import print_function
import random, os, sys, numpy, time
from spt3g import core

port = random.randint(10000, 60000)

frames = []
for i in range(0, 20):
	f = core.G3Frame()
	f['Sequence'] = i
	f['Data'] = core.G3Timestream(numpy.zeros(100000))
	frames.append(f)

print('Port: ', port)
child = os.fork()

if child != 0:
	# Parent
	print('Parent')	

	send = core.G3NetworkSender(hostname='*', port=port)
	time.sleep(1) # XXX: how to signal that the remote end is ready?
	print('Sending')

	for f in frames:
		send(f)
	send(core.G3Frame(core.G3FrameType.EndProcessing))

	pid, status = os.wait()
	print('Child Status: ', status)
	if status == 0:
		print('OK')
	sys.exit(status)
else:
	# Child
	print('Child')

	recv = core.G3Reader(filename='tcp://localhost:%d' % port)
	rframes = []
	for k in range(len(frames)):
		chunk = recv(None)
		print(chunk[0])
		rframes += chunk
	assert(len(rframes)) == len(frames)
	for i in range(len(rframes)):
		assert(rframes[i]['Sequence'] == i)

	sys.exit(0)



#!/usr/bin/env python

import numpy as np
from spt3g import core
import unittest

class TestBackAndForthConversion(unittest.TestCase):
    def test_complex_vec_conversion(self):
        c = np.arange(100000, dtype = 'complex128') + np.arange(100000, dtype = 'complex128')[::-1]*1j
        d = core.G3VectorComplexDouble(c)
        self.assertTrue(np.max( np.abs(c-d)) == 0)
        
        # Make sure item access works too; otherwise the above is essentially a test of memcpy()
        e = np.asarray(list(d))
        self.assertTrue(np.max( np.abs(c-e)) == 0)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from spt3g import core, maps
import numpy as np

np.random.seed(42)

n_samps = int(1e3)

az_0 = core.G3Timestream(np.random.rand(n_samps)*2.0*np.pi - np.pi)
pole_avoidance = 0.7
el_0 = core.G3Timestream(np.random.rand(n_samps)*np.pi*pole_avoidance-np.pi/2.0*pole_avoidance)

az_0.start = core.G3Time('20170329_000001')
el_0.start = core.G3Time('20170329_000001')

az_0.stop = core.G3Time('20170329_100001')
el_0.stop = core.G3Time('20170329_100001')

az_1 = az_0 + 10 * core.G3Units.arcmin
el_1 = el_0 + 10 * core.G3Units.arcmin

ra_0, dec_0 = maps.azel.convert_azel_to_radec(az_0, el_0)
ra_1, dec_1 = maps.azel.convert_azel_to_radec(az_1, el_1)

o_az_0 = az_0 + (np.random.rand()-0.5) * 2 * core.G3Units.deg
o_el_0 = el_0 + (np.random.rand()-0.5) * 2 * core.G3Units.deg
o_ra_0, o_dec_0 = maps.azel.convert_azel_to_radec(o_az_0, o_el_0)

import astropy.coordinates
from astropy import units as u
c = astropy.coordinates.FK5(ra=np.asarray(o_ra_0)*u.rad, dec=np.asarray(o_dec_0)*u.rad)
g = c.transform_to(astropy.coordinates.Galactic())
l_test = g.l.radian
b_test = g.b.radian


def sloppy_eq(f, g, eps=1e-5):
    np.testing.assert_allclose(f, g, rtol=0, atol=eps)

def apply_trans(alpha, delta, trans):
    q_off = maps.ang_to_quat(alpha, delta)
    ra, dec = maps.quat_to_ang(trans * q_off / trans)
    if ra < 0:
        ra += 2 * np.pi
    return ra, dec

for i in range(n_samps):
    q_trans = maps.get_transform_quat(
        az_0[i], -el_0[i], ra_0[i], dec_0[i],
        az_1[i], -el_1[i], ra_1[i], dec_1[i]
    );

    t_ra_0, t_dec_0 = apply_trans(az_0[i], -el_0[i], q_trans)
    sloppy_eq([t_ra_0, t_dec_0], [ra_0[i], dec_0[i]])

    t_ra_1, t_dec_1 = apply_trans(az_1[i], -el_1[i], q_trans)
    sloppy_eq([t_ra_1, t_dec_1], [ra_1[i], dec_1[i]])

    t_ra_o, t_dec_o = apply_trans(o_az_0[i], -o_el_0[i], q_trans)
    sloppy_eq([t_ra_o, t_dec_o], [o_ra_0[i], o_dec_0[i]])

    q_trans_gal = maps.get_fk5_j2000_to_gal_quat() * q_trans
    t_l_o, t_b_o = apply_trans(o_az_0[i], -o_el_0[i], q_trans_gal)
    sloppy_eq([t_l_o, t_b_o], [l_test[i], b_test[i]])



#!/usr/bin/env python

from spt3g import core
import unittest

import numpy as np
import copy

SEC = core.G3Units.sec



class TestTimesampleVector(unittest.TestCase):
    def test_from_list(self):
        t0 = core.G3Time('2019-01-01T12:30:00')
        vectime = core.G3VectorTime([t0, t0 + 10*SEC])
        assert(vectime[0].time == t0.time)
        assert(vectime[1].time == t0.time + 10*SEC)
    def test_from_numpy_array(self):
        t0 = core.G3Time('2019-01-01T12:30:00')
        timestamps = np.linspace(t0.time, t0.time + 1e7*SEC, 3000)
        vectime = core.G3VectorTime(timestamps)
        assert(vectime[0] == t0)
        assert(vectime[-1] == core.G3Time(timestamps[-1]))
        assert(len(vectime) == len(timestamps))
    def test_to_numpy_array(self):
        t0 = core.G3Time('2019-01-01T12:30:00')
        timestamps = np.linspace(t0.time, t0.time + 1e7*SEC, 3000, dtype='int64')
        vectime = core.G3VectorTime(timestamps)
        np.testing.assert_array_equal(vectime, timestamps)
    def test_reinit_from_numpy_array(self):
        t0 = core.G3Time('2019-01-01T12:30:00')
        timestamps = np.linspace(t0.time, t0.time + 1e7*SEC, 3000, dtype='int64')
        vectime = core.G3VectorTime(timestamps)
        revectime = core.G3VectorTime(np.asarray(vectime))
        np.testing.assert_array_equal(revectime, timestamps)
    def test_copy_constructor(self):
        t0 = core.G3VectorTime(np.array([100000000, 200000000]))
        t1 = core.G3VectorTime(t0)
        np.testing.assert_array_equal(t0, t1)


def get_test_block(length, keys=['a', 'b', 'c', 'd'],
                   offset=0, ordered=True):
    type_cycle = [(core.G3VectorDouble, float),
                  (core.G3VectorInt, int),
                  (core.G3VectorString, str),
                  (core.G3VectorBool, bool)]
    t0 = core.G3Time('2019-01-01T12:30:00') + offset*SEC
    m = core.G3TimesampleMap()
    times = np.arange(length)
    if not ordered:
        np.random.shuffle(times)
    m.times = core.G3VectorTime(t0 + times*SEC)
    for i, k in enumerate(keys):
        y = (np.random.uniform(size=length) * 100).astype(int)
        constructor, cast_func = type_cycle[i % len(type_cycle)]
        vect = constructor(list(map(cast_func, y)))
        m[k] = vect
    return m


class TestTimesampleMap(unittest.TestCase):
    def test_00_internal_checks(self):
        # Valid block.
        m = get_test_block(100)
        m.check()

    def test_10_safety(self):
        m0 = get_test_block(100)
        m1 = get_test_block(101)
        # Try to add an incompatible element.
        with self.assertRaises(ValueError):
            m0.times = m1.times
        with self.assertRaises(ValueError):
            m0['a'] = m1['d']
        # But we should be able to change times in an empty vector.
        m0 = get_test_block(100, [])
        m1 = get_test_block(101)
        m0.times = m1.times
        m0['x'] = m1['a']

    def test_20_concat(self):
        # Test concatenation.
        key_list = ['w', 'x', 'y', 'z']
        m0 = get_test_block(100, key_list)
        m1 = get_test_block(200, key_list, offset=100)
        m01 = m0.concatenate(m1)
        self.assertTrue(np.all(
            np.hstack([np.array(m0.times), np.array(m1.times)]) == np.array(m01.times)))
        for k in key_list:
            self.assertTrue(np.all(
                np.hstack([np.array(m0[k]), np.array(m1[k])]) == np.array(m01[k])))
        for fail_vec in [
                get_test_block(200, key_list + ['extra'], 100),
                get_test_block(200, key_list[:-1], 100),
                ]:
            with self.assertRaises(RuntimeError):
                m0.concatenate(fail_vec)

    def test_30_serialization(self):
        m0 = get_test_block(100)
        m1 = get_test_block(200, offset=100)
        m2 = m0.concatenate(m1)
        m0.check()
        m1.check()
        m2.check()
        f = core.G3Frame()
        f['irreg0'] = m0
        f['irreg1'] = m1
        core.G3Writer('test.g3').Process(f)
        f = core.G3Reader('test.g3').Process(None)[0]
        f['irreg0'].check()
        f['irreg1'].check()
        f['irreg0'].concatenate(f['irreg1'])['b']

    def test_40_sort(self):
        m0 = get_test_block(100, ordered=False)
        m1 = copy.deepcopy(m0)
        m0.sort()
        idx = np.argsort(m1.times)
        self.assertTrue((np.asarray(m1.times)[idx] == np.asarray(m0.times)).all())
        for k in m0.keys():
            self.assertTrue((np.asarray(m1[k])[idx] == np.asarray(m0[k])).all())


if __name__ == '__main__':
    unittest.main()



from . import G3Logger, G3LogLevel, usefulfunc
import traceback
from functools import reduce

__all__ = [
    "log_trace",
    "log_debug",
    "log_info",
    "log_notice",
    "log_warn",
    "log_error",
    "log_fatal",
    "set_log_level",
]


def log(log_level, unit, *args):
    message = reduce(lambda a,b: '%s %s' % ( str(a), str(b)), args)
    tb = traceback.extract_stack(limit=3)[0]
    G3Logger.global_logger.log(log_level, unit, tb[0], tb[1], tb[2], message)
    if log_level == G3LogLevel.LOG_FATAL:
        raise RuntimeError(message + " (in " + tb[2] + ")")


@usefulfunc
def log_trace(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_TRACE.
    """
    return log(G3LogLevel.LOG_TRACE, unit, *args)

@usefulfunc
def log_debug(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_DEBUG.
    """
    return log(G3LogLevel.LOG_DEBUG, unit, *args)

@usefulfunc
def log_info(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_INFO.
    """
    return log(G3LogLevel.LOG_INFO, unit, *args)

@usefulfunc
def log_notice(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_NOTICE.
    """
    return log(G3LogLevel.LOG_NOTICE, unit, *args)

@usefulfunc
def log_warn(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_WARN.
    """
    return log(G3LogLevel.LOG_WARN, unit, *args)

@usefulfunc
def log_error(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_ERROR.
    """
    return log(G3LogLevel.LOG_ERROR, unit, *args)

@usefulfunc
def log_fatal(*args, unit="Python"):
    """
    Log a message for the given unit at level LOG_FATAL,
    and raise a RuntimeError.
    """
    message, tb = log(G3LogLevel.LOG_FATAL, unit, *args)

@usefulfunc
def set_log_level(level, unit=None):
    '''
    Set log level to the requested level. If unit is not None, set the
    log level for the given logging unit only.
    
    Example
    -------
    ::

        core.set_log_level(core.G3LogLevel.LOG_DEBUG, 'GCPMuxDataDecoder')
    '''

    if unit is not None:
        G3Logger.global_logger.set_level_for_unit(unit, level)
    else:
        G3Logger.global_logger.set_level(level)


import atexit
def fix_logging_crash():
    # Unload any python loggers at exit to prevent Py_DECREF() after
    # interpreter destruction
    G3Logger.global_logger = None
atexit.register(fix_logging_crash)



from .. import core, calibration, dfmux
import numpy
from .MPIAccumulator import MPIAccumulator
import pickle

@core.indexmod
class TOASTFiller(object):
    '''
    Fill a TOAST TOD (stored in the toastobses member after completion) from
    a frame stream. Data will be distributed across the given MPI communicator
    in the way that TOAST would like.
    '''
    def __init__(self, mpicomm, boresight='OnlineRaDecRotation', timestreams='RawTimestreams_I', wiring='WiringMap'):
        self.boresight = boresight
        self.timestreams = timestreams
        self.wiring = wiring
        self.mpicomm = mpicomm
        self.mpia = MPIAccumulator(mpicomm, self.extractor, lambda m: m[0]['start'], dataframes=[core.G3FrameType.Scan])

    def extractor(self, frame):
        if frame.type == core.G3FrameType.Wiring:
            self.keys = frame[self.wiring].keys()
            self.invkeys = {k: i for i,k in enumerate(self.keys)}

        if frame.type == core.G3FrameType.Calibration:
            # from toast import qarray as qa
            #self.detquat = {k: qa.from_angles(v.y_offset/core.G3Units.rad, v.x_offset/core.G3Units.rad, 0) for k,v in frame['BolometerProperties'].items()}
            self.detquat = None

        if frame.type == core.G3FrameType.Scan:
            timestreams = frame[self.timestreams]
            bs = numpy.asarray(frame[self.boresight])
            bs = bs[:,[1,2,3,0]] # Swizzle for TOAST
            scan = {'dets': self.keys, 'boresight': bs, 'detquat': self.detquat, 'obsid': '%s-%d' % (frame['SourceName'], frame['ObservationID']), 'start': timestreams.start, 'stop': timestreams.stop, 'nsamples': len(timestreams.values()[0])}
            return scan['obsid'], scan, timestreams


    def __call__(self, frame):
        rv = self.mpia(frame)
        if frame.type != core.G3FrameType.EndProcessing:
            return rv

        from toast.tod import TODCache

        # Build TOAST TOD from our full data set, now that we have it
        toastobses = []
        for obs,scans in self.mpia.fullobs.items():
            nsamptot = sum([i[0]['nsamples'] for i in scans])
            d = TODCache(self.mpicomm, scans[0][0]['dets'], nsamptot, detquats=scans[0][0]['detquat'])

            t = numpy.concatenate([numpy.linspace(i[0]['start'].time/core.G3Units.s, i[0]['stop'].time/core.G3Units.s, i[0]['nsamples']) for i in scans])
            d.write_times(local_start=0, stamps=t[slice(*d.local_samples)])
            bs = numpy.concatenate([i[0]['boresight'] for i in scans])
            d.write_boresight(local_start=0, data=bs[slice(*d.local_samples)])

            startsample = 0
            for i in scans:
                i[0]['startsamp'] = startsample
                startsample += i[0]['nsamples']

            # Make a list of who needs what
            dataneeds = (self.mpicomm.rank, d.local_samples, d.local_dets)
            dataneeds = self.mpicomm.allgather(dataneeds)

            # Send/receive needed data
            outdata = [[] for i in range(self.mpicomm.size)]
            for need in dataneeds:
                for i in scans:
                    if i[2] is None:
                        continue
                    mask = slice(max(need[1][0], i[0]['startsamp']),
                            min(need[1][1] + need[1][0], i[0]['startsamp'] +
                                i[0]['nsamples']))
                    if mask.stop <= mask.start:
                        continue # No overlap

                    # Scan overlaps a need, get the relevant timestreams
                    maskstart = mask.start
                    mask = slice(mask.start - i[0]['startsamp'], mask.stop - i[0]['startsamp'])
                    chunk = {}
                    for k in need[2]:
                        chunk[k] = numpy.asarray(i[2][k])[mask]
                    outdata[need[0]].append((maskstart, chunk))
            del dataneeds

            outdata = self.mpicomm.alltoall(outdata) # Swap with everyone

            # Now stitch everything into the TOAST structure
            for sourcenode in outdata:
                for chunk in sourcenode:
                    localstart = chunk[0] - d.local_samples[0]
                    for k,v in chunk[1].items():
                        d.write(k, localstart, v)

            del self.mpia # Don't need this anymore
            toastobses.append({'id': obs, 'tod': d})

        self.toastobses = toastobses
        return frame




import numpy
import warnings
from . import G3SkyMapWeights, G3SkyMap, FlatSkyMap, G3SkyMapMask

__all__ = []

# This file adds extra functionality to the python interface to G3SkyMap and
# G3SkyMapWeights. This is done in ways that exploit a large fraction of
# the evil things you can do in Python (Nick referred to it as "devil magic"),
# but please believe that the alternative is worse. I have tried to explain
# the rationale for each block.

# Bind all numpy binary operators to G3SkyMap by stealing the numpy functions
# and wrapping them in short lambdas that cast to and from sky maps

def numpyinplace(skymap, array):
    # Some magic to give skymaps the same operators as numpy arrays. This
    # requires multiple lines, so cannot be in the lambda below.  Because
    # this calls __copy__(), which is Clone(), it does the right thing for
    # subclasses of skymaps.
    rv = skymap.__copy__()
    numpy.asarray(rv)[:] = array
    return rv

def numpycompat(a, b, op):
    if isinstance(b, G3SkyMap) and not a.compatible(b):
        raise TypeError("Map of type <{}> is incompatible with map of type <{}>".format(a, b))
    return numpyinplace(a, numpy.ndarray.__dict__[op](numpy.asarray(a), numpy.asarray(b)))


for x in ['__and__', '__divmod__', '__floordiv__', '__iand__', '__ifloordiv__', '__imod__', '__ior__', '__mod__', '__or__', '__rdivmod__', '__rmod__', '__rpow__']:
    setattr(G3SkyMap, x, lambda a, b, op=x: numpycompat(a, b, op))

for op in ["all", "any", "sum", "mean", "var", "std", "min", "max", "argmin", "argmax"]:
    def ufunc_wrapper(op):
        def ufunc(a, *args, **kwargs):
            bound_args = {}
            bound = getattr(a, "_c" + op)
            if op in ["std", "var", "nanstd", "nanvar"]:
                if "ddof" in kwargs:
                    bound_args["ddof"] = kwargs.pop("ddof")
            if "where" in kwargs:
                where = kwargs.pop("where")
                if isinstance(where, numpy.ndarray):
                    where = G3SkyMapMask(a, where)
                bound_args["where"] = where
            for k in ["axis", "out", "dtype"]:
                if k in kwargs and kwargs[k] is None:
                    kwargs.pop(k)
            if len(args) or len(kwargs):
                raise TypeError("ufunc {} does not support complex arguments".format(op))
            return bound(**bound_args)
        ufunc.__doc__ = getattr(numpy.ndarray, op).__doc__
        return ufunc
    setattr(G3SkyMap, op, ufunc_wrapper(op))
    if op in ["all", "any", "sum"]:
        setattr(G3SkyMapMask, op, ufunc_wrapper(op))

# Make weight maps so that you can index them and get the full 3x3 weight matrix

def skymapweights_keys(self):
    """
    Return the list of string names of valid weight attributes.  For unpolarized
    weights, this list includes only TT.  Otherwise, the list includes all six
    unique weight attributes in row major order: TT, TQ, TU, QQ, QU, UU.
    """
    if self.polarized:
        return ['TT', 'TQ', 'TU', 'QQ', 'QU', 'UU']
    return ['TT']

G3SkyMapWeights.keys = skymapweights_keys

def skymapweights_getitem(self, x):
    if isinstance(x, str) and x in self.keys():
        return getattr(self, x)

    if (isinstance(x, tuple) and any(isinstance(xx, slice) for xx in x)) \
       or isinstance(x, slice):
        out = self.__class__()
        out.TT = self.TT[x]
        if self.polarized:
            out.TQ = self.TQ[x]
            out.TU = self.TU[x]
            out.QQ = self.QQ[x]
            out.QU = self.QU[x]
            out.UU = self.UU[x]
        return out

    if not self.polarized:
        return self.TT[x]

    mat = numpy.zeros((3,3))
    mat[0,0] = self.TT[x]
    mat[0,1] = mat[1,0] = self.TQ[x]
    mat[0,2] = mat[2,0] = self.TU[x]
    mat[1,1] = self.QQ[x]
    mat[1,2] = mat[2,1] = self.QU[x]
    mat[2,2] = self.UU[x]

    return mat

G3SkyMapWeights.__getitem__ = skymapweights_getitem

def skymapweights_setitem(self, x, mat):
    if isinstance(x, str) and x in self.keys():
        setattr(self, x, mat)
        return

    if (isinstance(x, tuple) and any(isinstance(xx, slice) for xx in x)) \
       or isinstance(x, slice):
        raise NotImplementedError

    if not self.polarized:
        assert(numpy.isscalar(mat))
        self.TT[x] = mat
        return

    mat = numpy.asarray(mat, dtype=float)
    assert(mat.shape == (3, 3))
    assert(numpy.allclose(mat, mat.T))
    self.TT[x] = mat[0,0]
    self.TQ[x] = mat[0,1]
    self.TU[x] = mat[0,2]
    self.QQ[x] = mat[1,1]
    self.QU[x] = mat[1,2]
    self.UU[x] = mat[2,2]

G3SkyMapWeights.__setitem__ = skymapweights_setitem

# patch handling for flat sky maps
def skymapweights_extract(self, x0, y0, width, height, fill=0):
    if not isinstance(self.TT, FlatSkyMap):
        raise NotImplementedError

    w = self.__class__()
    for k in self.keys():
        setattr(w, k, self[k].extract_patch(x0, y0, width, height, fill=fill))

    return w

skymapweights_extract.__doc__ = FlatSkyMap.extract_patch.__doc__
G3SkyMapWeights.extract_patch = skymapweights_extract

def skymapweights_insert(self, weights_patch, ignore_zeros=False):
    if not isinstance(self.TT, FlatSkyMap):
        raise NotImplementedError

    assert weights_patch.polarized == self.polarized, \
        "Input weights patch must be {}polarized".format("" if self.polarized else "un")

    for k in self.keys():
        self[k].insert_patch(weights_patch[k], ignore_zeros=ignore_zeros)

skymapweights_insert.__doc__ = FlatSkyMap.insert_patch.__doc__
G3SkyMapWeights.insert_patch = skymapweights_insert

def skymapweights_reshape(self, width, height, fill=0):
    if not isinstance(self.TT, FlatSkyMap):
        raise NotImplementedError

    w = self.__class__()
    for k in self.keys():
        setattr(w, k, self[k].reshape(width, height, fill=fill))

    return w

skymapweights_reshape.__doc__ = FlatSkyMap.reshape.__doc__
G3SkyMapWeights.reshape = skymapweights_reshape

# Pass through attributes to submaps. This is not ideal because the properties
# are hidden and not visible by tab completion. A better solution would use
# __getattribute__
def skymapweights_getattr(self, x):
    if x == "flat_pol" and isinstance(self.TQ, FlatSkyMap):
        return getattr(self.TQ, x)
    if x == "pol_conv" and isinstance(self.TU, FlatSkyMap):
        return getattr(self.TU, x)
    if hasattr(self.TT, x):
        return getattr(self.TT, x)
    raise AttributeError("'{}' object has no attribute '{}'".format(type(self), x))
G3SkyMapWeights.__getattr__ = skymapweights_getattr
oldsetattr = G3SkyMapWeights.__setattr__
def skymapweights_setattr(self, x, val):
    try:
        oldsetattr(self, x, val)
    except AttributeError:
        if self.TT is not None:
            setattr(self.TT, x, val)
        if self.TQ is not None:
            setattr(self.TQ, x, val)
        if self.TU is not None:
            setattr(self.TU, x, val)
        if self.QQ is not None:
            setattr(self.QQ, x, val)
        if self.QU is not None:
            setattr(self.QU, x, val)
        if self.UU is not None:
            setattr(self.UU, x, val)
G3SkyMapWeights.__setattr__ = skymapweights_setattr

def skymapmask_getattr(self, x):
    if hasattr(self.parent, x):
        return getattr(self.parent, x)
    raise AttributeError("'{}' object has no attribute '{}'".format(type(self), x))
G3SkyMapMask.__getattr__ = skymapmask_getattr

def skymap_clone(self, copy_data=True):
    with warnings.catch_warnings():
        warnings.simplefilter("default")
        warnings.warn(
            "{0}.Clone is deprecated, use {0}.clone instead".format(
                self.__class__.__name__
            ),
            DeprecationWarning,
        )
    return self.clone(copy_data)
G3SkyMap.Clone = skymap_clone
G3SkyMapWeights.Clone = skymap_clone

def skymap_compat(self, other):
    with warnings.catch_warnings():
        warnings.simplefilter("default")
        warnings.warn(
            "{0}.IsCompatible is deprecated, use {0}.compatible instead".format(
                self.__class__.__name__
            ),
            DeprecationWarning,
        )
    return self.compatible(other)
G3SkyMap.IsCompatible = skymap_compat



#!/usr/bin/env python

from spt3g import core
import time, sys

# File to disk
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
n = 0
def addinfo(fr):
	if fr.type != core.G3FrameType.Timepoint:
		return
	global n
	fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
	fr['count'] = n
	fr2 = core.G3Frame(core.G3FrameType.Housekeeping)
	fr2['count'] = n
	n += 1
	return [fr, fr2]
pipe.Add(addinfo)
pipe.Add(core.Dump)
pipe.Add(core.G3Writer, filename='test.g3')
pipe.Add(core.G3Writer, filename='test-hk.g3', streams=[core.G3FrameType.Housekeeping])
pipe.Run()

# And back from disk
print('Reading')
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename='test-hk.g3')
pipe.Add(core.Dump)
n = 0
def checkinfo(fr):
	global n
	if fr.type == core.G3FrameType.EndProcessing:
		return
	if fr.type != core.G3FrameType.Housekeeping:
		return
	if fr['count'] != n:
		print('Out of order frame')
		sys.exit(1)
	n += 1
pipe.Add(checkinfo)
pipe.Run()

if n != 10:
	print('Wrong number of frames (%d should be %d)' % (n, 10))
	sys.exit(1)

sys.exit(0)




from spt3g import core
import time, sys

# Read in one or more files specified at the command line and print their
# contents to the console.

pipe = core.G3Pipeline()

pipe.Add(core.G3Reader, filename=sys.argv[1:])
def printfr(fr):
	print(fr)
pipe.Add(printfr) # core.Dump does the same thing as this

pipe.Run()



#!/usr/bin/env python

from spt3g import core
import os, sys

# Test that we can read files written on a variety of platforms. Pass a path
# to generate test data for whatever this platform is.

testpath = os.path.join(os.path.dirname(__file__), "portability")

# Test data. Exercise some complicated things (STL bits) that we don't
# necessarily have control over, mapping a few primitive types.
f = core.G3Frame()
f['Five'] = 5
v = core.G3VectorDouble([2.6, 7.2])
f['Vec'] = v
v = core.G3VectorInt([17, 42, 87])
f['VecInt'] = v
m = core.G3MapDouble()
m['Six'] = 6
m['GoingOnSixteen'] = 15.9
f['Map'] = m

if len(sys.argv) > 1:
	core.G3Writer(sys.argv[1])(f)
	sys.exit(0)

# For now, we test files from big-endian (PPC64) and little-endian (amd64)
# 64-bit systems. Should include some 32-bit ones.

for test in ['test-be.g3', 'test-le.g3', 'test-le-v2.g3']:
	print(test)
	testdata = core.G3Reader(os.path.join(testpath, test))(None)[0]

	assert(testdata['Five'] == f['Five'])
	assert(len(testdata['Vec']) == len(f['Vec']))
	for i in range(len(testdata['Vec'])):
		assert(testdata['Vec'][i] == f['Vec'][i])
	assert(len(testdata['VecInt']) == len(f['VecInt']))
	for i in range(len(testdata['VecInt'])):
		assert(testdata['VecInt'][i] == f['VecInt'][i])
	assert(len(testdata['Map']) == len(f['Map']))
	assert(list(testdata['Map'].keys()) == list(f['Map'].keys()))
	for i in testdata['Map'].keys():
		assert(testdata['Map'][i] == f['Map'][i])
	




import numpy as np
import copy
from .. import core
from . import ACUStatus, ACUState, TrackerStatus, TrackerState, TrackerPointing, CalFile

def UnitValue(caldict_entry):
    '''Turn unit name into floating point unit value'''

    if 'UnitValue' in caldict_entry:
        return caldict_entry['UnitValue']

    try: 
        uname = caldict_entry['UnitName']
        if uname and uname != 'None':
            try:
                if '/' in uname:
                    unames = list(filter(None,uname.split('/')))
                    uvalue1 = getattr(core.G3Units, 
                                      list(filter(None,unames[0].split(' ')))[0])
                    uvalue2 = getattr(core.G3Units, 
                                      list(filter(None,unames[1].split(' ')))[0])
                    uvalue = uvalue1 / uvalue2
                else:
                    uvalue = getattr(core.G3Units, uname)
            except AttributeError:
                uvalue = 1.
                core.log_warn('No entry in G3Units for ' + uname + '. Setting UnitValue to 1.0\n')
        else:
            uvalue = 1.
    except KeyError:
        uvalue = 1.

    caldict_entry['UnitValue'] = uvalue

    return uvalue


def CalibrateValue(data, caldict_entry):
    '''Apply gain / offset units from G3 cal file to register'''

    uvalue = UnitValue(caldict_entry)

    # if a register has units, it can't be an int anymore.  well, actually,
    # it can't be an int if we're adding floats to it or multiplying it by
    # floats either, so convert everything that has an entry in the cal file
    # to float/double.
    if 'OutputType' not in caldict_entry:
        g3type = type(data)
        if g3type == core.G3VectorInt:
            g3type = core.G3VectorDouble
        elif g3type == core.G3MapInt:
            g3type = core.G3MapDouble
        elif g3type == core.G3Int:
            g3type = core.G3Double
        caldict_entry['OutputType'] = g3type
    g3type = caldict_entry['OutputType']

    # make a copy
    if np.size(data) == 1:
        data = data.value
    data2 = np.asarray(data, dtype='float64')

    # calibrate units
    offset, recip = caldict_entry['Offset'], caldict_entry['ReciprocalFactor']
    if offset != 0:
        data2 += float(offset)
    data2 *= float(uvalue / recip)
    if not data2.shape:
        data2 = data2.tolist()

    return g3type(data2)


@core.indexmod
class CalibrateFrame:
    '''Apply gain / offset / units from G3 cal file'''

    def __init__(self, calibration_file=None):
        cf = CalFile.CalFileReader()
        self.cal = cf.readCalFile(calibration_file)

    def __call__(self, frame):

        if frame.type != core.G3FrameType.GcpSlow or 'Calibrated' in frame:
            return

        for board in list(frame.keys()):
            cboard = frame.pop(board)
            if board not in self.cal:
                continue
            bcal = self.cal[board]
            for rmap, crmap in cboard.items():
                if rmap not in bcal:
                    continue
                rcal = bcal[rmap]
                for reg, creg in crmap.items():
                    if reg not in rcal:
                        continue
                    rcd = rcal[reg]
                    rsize = np.size(creg)
                    rshape = np.shape(creg)
                    if rsize > 1 and len(rshape) > 1:
                        for i in range(rshape[0]):
                            try:
                                rcdi = rcd[i]
                            except KeyError:
                                rcdi = rcd
                            creg[i] = CalibrateValue(creg[i], rcdi)
                    else:
                        try:
                            rcdi = rcd[0]
                        except KeyError:
                            rcdi = rcd
                        crmap[reg] = CalibrateValue(creg, rcdi)
            frame[board] = cboard

        frame['Calibrated'] = True


@core.indexmod
def UnpackACUData(f):
    '''Extracts ACU status information to ACUStatus key in frame'''

    if f.type != core.G3FrameType.GcpSlow:
        return

    board = f['antenna0']['acu']

    a = ACUStatus()
    a.time = f['antenna0']['frame']['utc']
    a.az_pos = board['az_pos'].value
    a.el_pos = board['el_pos'].value
    a.az_rate = board['az_rate'].value
    a.el_rate = board['el_rate'].value

    # 'new_*' registers not actually filled by GCP; ignore them

    a.px_checksum_error_count = board['px_checksum_error_count'].value
    a.px_resync_count = board['px_resync_count'].value
    a.px_resync_timeout_count = board['px_resync_timeout_count'].value
    a.px_resyncing = board['px_resyncing'].value
    a.px_timeout_count = board['px_timeout_count'].value
    a.restart_count = board['restart_count'].value

    a.state = ACUState(board['state'].value)
    a.status = board['acu_status'].value
    try:
        a.error = board['acu_error'].value
    except KeyError:
        # This register was some time in early 2018.  In order to read
        # older data, just set the error code to 0.
        a.error = 0

    f['ACUStatus'] = a


@core.indexmod
def UnpackTrackerMinimal(f, rewrite_source_from_feature_bits=True):
    '''
    Construct SourceName and ObservationId keys from frame.

    If rewrite_source_from_feature_bits is True (the default), will try to
    rewrite source names if DecryptFeatureBit() has been run and either
    "elnod", "calibrator", or "noise" is present in the feature bit list
    to that value.
    '''

    if f.type != core.G3FrameType.GcpSlow:
        return

    # Grab the GCP source name. If it is "current", fill in something more
    # helpful from the feature bits if possible.
    source = f['antenna0']['tracker']['source'].value
    if rewrite_source_from_feature_bits and 'GCPFeatureBits' in f:
        if 'elnod' in f['GCPFeatureBits']:
            source = 'elnod'
        if 'calibrator' in f['GCPFeatureBits'] and 'source_scan' not in f['GCPFeatureBits']:
            source = 'calibrator'
        if 'noise' in f['GCPFeatureBits']:
            source = 'noise'
        if 'debug' in f['GCPFeatureBits']:
            source = 'debug-forced-scanify'
        if 'every_pixel_on_src' in f['GCPFeatureBits']:
            source = source + '-pixelraster' # NB: Do NOT use in-place +=
    f['SourceName'] = source

    # And observation ID, if present
    if 'obs_id' in f['antenna0']['tracker']:
        f['ObservationID'] = f['antenna0']['tracker']['obs_id']


@core.indexmod
def UnpackTrackerData(f, rewrite_source_from_feature_bits=True):
    '''
    Extracts tracker status information to frame into the TrackerStatus key,
    along with the observation processing handled by UnpackTrackerMinimal.

    If rewrite_source_from_feature_bits is True (the default), will try to
    rewrite source names if DecryptFeatureBit() has been run and either
    "elnod", "calibrator", or "noise" is present in the feature bit list
    to that value.
    '''

    if f.type != core.G3FrameType.GcpSlow:
        return

    UnpackTrackerMinimal(f, rewrite_source_from_feature_bits)

    board = f['antenna0']['tracker']

    t = TrackerStatus()
    # List comprehensions are due to funny business with G3VectorFrameObject
    t.time = board['utc'][0]

    # Measured values
    t.az_pos = np.asarray(board['actual'][0])
    t.el_pos = np.asarray(board['actual'][1])
    # XXX units for rates seem to be wrong. I think this is in encoder counts
    t.az_rate = np.asarray(board['actual_rates'][0], dtype = float)
    t.el_rate = np.asarray(board['actual_rates'][1], dtype = float)
    
    # Expected values
    t.az_command = np.asarray(board['expected'][0])
    t.el_command = np.asarray(board['expected'][1])
    t.az_rate_command = np.asarray(board['expected_rates'][0], dtype = float)
    t.el_rate_command = np.asarray(board['expected_rates'][1], dtype = float)

    # Status params
    if isinstance(board['state'][0], core.G3String):
        # If state is all zero (LACKING), for example due to an ACU glitch,
        # the ARC reader may decide that the 8-bit array field is a string.
        # Treat it as one.
        t.state = [TrackerState(0) for s in board['inControl'][0]]
    else:
        t.state = [TrackerState(s) for s in board['state'][0]]
    t.acu_seq = board['acu_seq'][0]
    t.in_control = core.BoolVector(board['inControl'][0])
    t.in_control_int = core.IntVector(board['inControl'][0])
    t.scan_flag = core.BoolVector(board['scan_flag'][0])
    
    t.lst = np.asarray(board['lst'][0], dtype=float)

    t.source_acquired = np.asarray(board['off_source'][0])
    t.source_acquired_threshold = np.asarray(board['source_acquired_threshold'])
    t.tracker_mode = np.asarray(board['mode'][0])
    t.tracker_lacking = np.asarray(board['lacking'][0])
    t.time_status = np.asarray(board['time_status'][0])
    try:
        t.schedule_name = np.asarray(board['schedule_name'].value)
    except AttributeError:
        t.schedule_name = np.asarray(''.join([chr(x) for x in board['schedule_name']]))

    f['TrackerStatus'] = t


@core.indexmod
def UnpackTrackerPointingData(f):
    '''
    Extracts tracker registers relevant to online and offline pointing.
    Calibration values (offsets and multiplicative constants) are from
    gcp/control/conf/spt/cal.
    '''

    if f.type != core.G3FrameType.GcpSlow:
        return

    board = f['antenna0']['tracker']

    t = TrackerPointing()
    t.time = board['utc'][0]
    t.scu_temp = np.asarray(f['antenna0']['scu']['temp'])
    t.features = core.IntVector([f['array']['frame']['features'].value])

    t.encoder_off_x = np.asarray([board['encoder_off'][0]], dtype=np.double)
    t.encoder_off_y = np.asarray([board['encoder_off'][1]], dtype=np.double)
    
    t.low_limit_az = np.asarray([board['az_limits'][0]], dtype=np.double)
    t.high_limit_az = np.asarray([board['az_limits'][1]], dtype=np.double)
    t.low_limit_el = np.asarray([board['el_limits'][0]], dtype=np.double)
    t.high_limit_el = np.asarray([board['el_limits'][1]], dtype=np.double)

    t.tilts_x = np.asarray(board['tilt_xy_avg'][0], dtype=np.double)
    t.tilts_y = np.asarray(board['tilt_xy_avg'][1], dtype=np.double)
    t.refraction = np.asarray(board['refraction'][2], dtype=np.double)

    t.horiz_mount_x = np.asarray(board['horiz_mount'][0])
    t.horiz_mount_y = np.asarray(board['horiz_mount'][1])
    t.horiz_off_x = np.asarray(board['horiz_off'][0])
    t.horiz_off_y = np.asarray(board['horiz_off'][1])

    t.scan_off_x = np.asarray(board['scan_off'][0])
    t.scan_off_y = np.asarray(board['scan_off'][1])
    t.sky_off_x = np.asarray(board['sky_xy_off'][0])
    t.sky_off_y = np.asarray(board['sky_xy_off'][1])
    t.equat_off_x = np.asarray(board['equat_off'][0])
    t.equat_off_y = np.asarray(board['equat_off'][1])

    t.equat_geoc_ra = np.asarray(board['equat_geoc'][0])
    t.equat_geoc_dec = np.asarray(board['equat_geoc'][1])
    t.horiz_topo_az = np.asarray(board['horiz_topo'][0])
    t.horiz_topo_el = np.asarray(board['horiz_topo'][1])

    t.error_az = np.asarray(board['errors'][0])
    t.error_el = np.asarray(board['errors'][1])

    t.linsens_avg_l1 = np.asarray(board['linear_sensor_avg'][0])
    t.linsens_avg_l2 = np.asarray(board['linear_sensor_avg'][1])
    t.linsens_avg_r1 = np.asarray(board['linear_sensor_avg'][2])
    t.linsens_avg_r2 = np.asarray(board['linear_sensor_avg'][3])
    
    t.telescope_temp = np.asarray([f['array']['weather']['airTemperature'].value])
    t.telescope_pressure = np.asarray([f['array']['weather']['pressure'].value])

    f['TrackerPointing'] = t

    p = core.G3MapVectorDouble()
    pc = core.G3MapVectorDouble()

    # core model parameters
    for k in ["tilts", "flexure", "fixedCollimation"]:
        p[k] = np.asarray(board[k], dtype=np.double)
        kadj = k + "Adjust"
        if kadj in board.keys():
            v = np.asarray(board[kadj], dtype=np.double)
            if np.any(v):
                pc[k] = v

    p['time'] = np.asarray(t.time, dtype=np.double)
    if 'linear_sensor_enabled' in board.keys():
        p['linsensEnabled'] = np.asarray(board['linear_sensor_enabled'][0], dtype=np.double)
        if 'linear_sensor_coeff' in board.keys():
            p['linsensCoeffs'] = np.asarray(board['linear_sensor_coeff'], dtype=np.double)
    else:
        p['linsensEnabled'] = np.zeros_like(p['time'], dtype=np.double)

    f['OnlinePointingModel'] = p
    if len(pc.keys()):
        pc['time'] = p['time']
        f['OnlinePointingModelCorrection'] = pc


@core.indexmod
def UpdateLinearSensorDeltas(f):

    if 'TrackerPointing' not in f:
        return

    if 'LinearSensorDeltas' in f:
        return

    # Yoke dimensions in mm.
    Rs = 1652.0 * core.G3Units.mm
    Rh = 3556.0 * core.G3Units.mm
    Ry = 6782.0 * core.G3Units.mm

    l1 = f["TrackerPointing"].linsens_avg_l1
    l2 = f["TrackerPointing"].linsens_avg_l2
    r1 = f["TrackerPointing"].linsens_avg_r1
    r2 = f["TrackerPointing"].linsens_avg_r2

    # Calculate corrections in radians.
    p = core.G3MapVectorDouble()
    p['time'] = np.asarray(f['TrackerPointing'].time, dtype=np.double)
    p["delta_az"] = (Rh / (Ry * Rs)) * (l1 - l2 - r1 + r2) * core.G3Units.rad
    p["delta_el"] = (1. / (2. * Rs)) * (l2 - l1 + r2 - r1) * core.G3Units.rad
    p["delta_et"] = (1. / (2. * Ry)) * (l1 + l2 - r1 - r2) * core.G3Units.rad

    f['LinearSensorDeltas'] = p


@core.indexmod
def DecryptFeatureBit(f):
    '''
    Unpacks the GCP feature flags
    '''

    if f.type != core.G3FrameType.GcpSlow:
        return

    flag_array = core.G3VectorString()
    feature_bit = f['array']['frame']['features'].value

    flags = ['analyze', 'source_scan', 'cabin_shutter', 'elnod', 'pol_cal',
             'calibrator', 'every_pixel_on_src', 'skydip', 'optical', 'noise',
             'trail', 'el_scan', None, None, None, None, None, None, None,
             'debug']
             # Sorry... NDH

    for i in enumerate(flags):
        if feature_bit & (1 << i[0]):
            if i[1] is None:
                core.log_error('Got an unused feature bit: {:d}'.format(i[0]))
            flag_array.append(i[1])

    f['GCPFeatureBits'] = flag_array


@core.indexmod
def AddBenchData(f):
    '''
    Add the optical bench positions to the frame.
    '''
    if f.type != core.G3FrameType.GcpSlow:
        return
    bench_axes = ['y1', 'y2', 'y3', 'x4', 'x5', 'z6']

    board = f['antenna0']['scu']

    # As of 2017-08-03, SCU time is not trustworthy
    # time = board['benchSampleTime'][0]
    # For now, do this bit of evil
    time = f['antenna0']['tracker']['utc'][0]
    start = time[0]
    stop = time[-1]

    benchcom = core.G3TimestreamMap()
    benchpos = core.G3TimestreamMap()
    benchzero = core.G3TimestreamMap()
    benchoff = core.G3TimestreamMap()
    bencherr = core.G3TimestreamMap()
    bench_info = core.G3TimestreamMap()
    for i, key in enumerate(bench_axes):
        benchcom[key] = core.G3Timestream(board['benchExpected'][i])
        benchcom[key].start = start
        benchcom[key].stop = stop

        benchpos[key] = core.G3Timestream(board['benchActual'][i])
        benchpos[key].start = start
        benchpos[key].stop = stop

        benchzero[key] = core.G3Timestream(board['benchZeros'][i])
        benchzero[key].start = start
        benchzero[key].stop = stop

        benchoff[key] = core.G3Timestream(board['benchOffsets'][i])
        benchoff[key].start = start
        benchoff[key].stop = stop

        bencherr[key] = core.G3Timestream(board['benchErrors'][i])
        bencherr[key].start = start
        bencherr[key].stop = stop

    info_items = ['benchFocus', 'benchDeadBand', 'benchAcquiredThreshold',
                  'benchPrimaryState', 'benchSecondaryState', 
                  'benchFault', 'timeLocked']
    bench_info = core.G3TimestreamMap()
    for i, key in enumerate(info_items):
        bench_info[key] = core.G3Timestream(board[key][0])
        bench_info[key].start = start
        bench_info[key].stop = stop

    f['BenchPosition'] = benchpos
    f['BenchCommandedPosition'] = benchcom
    f['BenchZeros'] = benchzero
    f['BenchOffsets'] = benchoff
    f['BenchErrors'] = bencherr
    f['BenchInfo'] = bench_info
    f['BenchSampleTime'] = board['benchSampleTime'][0]
    
@core.indexmod
def UnpackCryoData(f):
    '''
    Extracts cryo information into CryoStatus key
    '''

    if f.type != core.G3FrameType.GcpSlow:
        return

    if 'cryo' not in f['array']:
        return

    board = f['array']['cryo']
    temps = board['temperature']
    heaters = board['heater_dac']

    t = core.G3MapDouble()
    t['cryo_is_valid'] = board['cryoIsValid'][0]

    # Measured values
    # He10
    t['uc_head'] = temps[0][0]
    t['ic_head'] = temps[0][1]
    t['he4_head'] = temps[0][2]
    t['he4_fb'] = temps[0][3]
    t['he4_pump'] = temps[0][4]
    t['ic_pump'] = temps[0][5]
    t['uc_pump'] = temps[0][6]
    t['he4_sw'] = temps[0][7]
    t['ic_sw'] = temps[0][8]
    t['uc_sw'] = temps[0][9]
    t['uc_stage'] = temps[0][10]
    t['lc_tower'] = temps[0][11]
    t['ic_stage'] = temps[0][12]
    t['t4k_head'] = temps[0][13]
    t['t4k_squid_strap'] = temps[0][14]
    t['t50k_head'] = temps[0][15]

    # Optics
    t['b1_50k_wbp_near'] = temps[1][0]
    t['b2_50k_wbp_far'] = temps[1][1]
    t['b3_50k_diving_board'] = temps[1][2]
    t['b4_50k_top_bot_ptc'] = temps[1][3]
    t['y1_50k_head'] = temps[1][4]
    t['y2_50k_window_strap_near'] = temps[1][5]
    t['y3_50k_tube_strap_near'] = temps[1][6]
    t['y4_50k_tube'] = temps[1][7]
    t['g1_4k_head'] = temps[1][8]
    t['g2_4k_strap'] = temps[1][9]
    t['g3_4k_lens_tab'] = temps[1][10]
    t['g4_4k_lens_tab_far'] = temps[1][11]
    t['r1_4k_top_top_ptc'] = temps[1][12]
    t['r2_50k_midop_bot_ptc'] = temps[1][13]
    t['r3_4k_lyot_flange'] = temps[1][14]
    t['r4_4k_lyot'] = temps[1][15]

    # Receiver
    t['t4k_plate_far'] = temps[2][0]
    t['t4k_strap_optics'] = temps[2][1]
    t['t4k_plate_mid'] = temps[2][2]
    t['t4k_plate_top'] = temps[2][3]
    t['t4k_plate_ptc'] = temps[2][4]
    t['t50k_harness_middle'] = temps[2][6]
    t['t50k_strap'] = temps[2][7]
    t['squid_wh1_sl1'] = temps[2][8]
    t['squid_wh5_sl1'] = temps[2][9]
    t['squid_wh3_sl7'] = temps[2][10]
    t['cal_filament'] = temps[2][11]
    t['cal_ambient1'] = temps[2][12]
    t['cal_ambient2'] = temps[2][13]
    t['cal_ambient3'] = temps[2][14]

    # Heaters
    t['heat_he4_pump'] = heaters[0][3]
    t['heat_ic_pump'] = heaters[0][4]
    t['heat_uc_pump'] = heaters[0][5]
    t['heat_he4_sw'] = heaters[0][0]
    t['heat_ic_sw'] = heaters[0][1]
    t['heat_uc_sw'] = heaters[0][2]

    f['CryoStatus'] = t
    f['CryoStatusTime'] = board['utc']


@core.indexmod
def UnpackPTData(f):
    '''Extracts pulse tube status information to PTStatus key 
    in frame'''

    if f.type != core.G3FrameType.GcpSlow:
        return

    if 'pt415' not in f['array']:
        return

    board = f['array']['pt415']

    p = core.G3MapDouble()

    p['optics_lowp'] = board['pressure_low'][0]
    p['min_optics_lowp'] = board['min_pressure_low'][0]
    p['max_optics_lowp'] = board['max_pressure_low'][0]
    p['optics_highp'] = board['pressure_high'][0]
    p['min_optics_highp'] = board['min_pressure_high'][0]
    p['max_optics_highp'] = board['max_pressure_high'][0]
    p['optics_tempoil'] = board['temp_oil'][0]
    p['min_optics_tempoil'] = board['min_temp_oil'][0]
    p['max_optics_tempoil'] = board['max_temp_oil'][0]

    p['receiver_lowp'] = board['pressure_low'][1]
    p['min_receiver_lowp'] = board['min_pressure_low'][1]
    p['max_receiver_lowp'] = board['max_pressure_low'][1]
    p['receiver_highp'] = board['pressure_high'][1]
    p['min_receiver_highp'] = board['min_pressure_high'][1]
    p['max_receiver_highp'] = board['max_pressure_high'][1]
    p['receiver_tempoil'] = board['temp_oil'][1]
    p['min_receiver_tempoil'] = board['min_temp_oil'][1]
    p['max_receiver_tempoil'] = board['max_temp_oil'][1]

    p['optics_is_valid'] = board['deviceIsValid'][0]
    p['receiver_is_valid'] = board['deviceIsValid'][1]

    f['PTStatus'] = p
    f['PTStatusTime'] = board['utc']


@core.indexmod
def UnpackMuxData(f):
    '''
    Add the DFMux data to the frame.
    '''
    if f.type != core.G3FrameType.GcpSlow:
        return

    try:
        mux = f['array']['muxHousekeeping']
        boards = mux['boardname']
    except KeyError:
        return

    fpga_temp = core.G3MapDouble()
    board_name = core.G3MapString()

    for i, bn in enumerate(boards):
        bn = str(bn).replace('"', '') # get rid of extra quotes in board name
        if bn != "":
            board_name[str(i)] = bn
            fpga_temp[str(i)] = mux['MB_TEMPERATURE_FPGA_DIE'][i]
    f['MuxFPGATemp'] = fpga_temp
    f['MuxBoardName'] = board_name
    f['MuxTime'] = mux['utc']


@core.indexmod
def UnpackWeatherData(f):
    '''Extracts weather status information to Weather key 
    in frame'''

    if f.type != core.G3FrameType.GcpSlow:
        return

    if 'weather' not in f['array']:
        return

    board = f['array']['weather']

    t = core.G3MapDouble()
    t['telescope_temp'] = board['airTemperature'].value
    t['telescope_pressure'] = board['pressure'].value
    t['inside_dsl_temp'] = board['internalTemperature'].value
    t['wind_speed'] = board['windSpeed'].value
    t['wind_direction'] = board['windDirection'].value
    t['battery'] = board['battery'].value
    t['rel_humidity'] = board['relativeHumidity'].value
    t['power'] = board['power'].value
    t['tau'] = f['array']['tipper']['tau'].value
    t['tatm'] = f['array']['tipper']['tatm'].value

    f['Weather'] = t
    f['WeatherTime'] = board['utc']
    f['TipperTime'] = f['array']['tipper']['utc']


@core.pipesegment
def ARCExtract(pipe):
    '''Extract GCP registers into native objects'''
    pipe.Add(CalibrateFrame)
    pipe.Add(UnpackACUData)
    pipe.Add(UnpackTrackerPointingData)
    pipe.Add(UpdateLinearSensorDeltas)
    pipe.Add(DecryptFeatureBit)
    pipe.Add(UnpackTrackerData)
    pipe.Add(AddBenchData)
    pipe.Add(UnpackCryoData)
    pipe.Add(UnpackPTData)
    pipe.Add(UnpackMuxData)
    pipe.Add(UnpackWeatherData)


@core.pipesegment
def ARCExtractMinimal(pipe):
    '''
    Extract bare minimum GCP registers into native objects.

    Includes only GCPFeatureBits, SourceName and ObservationID keys.
    Use ARCExtract to calibrate and extract the complete frame.
    '''
    pipe.Add(DecryptFeatureBit)
    pipe.Add(UnpackTrackerMinimal)

# Need tool for tilt meter next



#!/usr/bin/env python
from spt3g import core, calibration

# Test that parts of BolometerPropertiesMap are appropriately settable

a = calibration.BolometerProperties()
a.pol_angle = 3
a.pol_angle += 3
assert(a.pol_angle == 6)
b = calibration.BolometerPropertiesMap()
b['test'] = a
assert(b['test'].pol_angle == 6)
b['test'].pol_angle += 6
print(b['test'].pol_angle)
assert(b['test'].pol_angle == 12)




#!/usr/bin/env python

from spt3g.core import G3Bool
import unittest

class TestDatatypes(unittest.TestCase):
    def test_g3bool(self):
        t = G3Bool(True)
        f = G3Bool(False)
        self.assertTrue(t and t.value)
        self.assertFalse(f or f.value)

if __name__ == '__main__':
    unittest.main()



from .. import core
from ..core import G3TimestreamQuat
from . import MapCoordReference
from . import get_origin_rotator_timestream, get_boresight_rotator_timestream
from . import get_fk5_j2000_to_gal_quat


__all__ = [
    "FillCoordTransRotations",
    "EquatorialToGalacticTransRotations",
    "AddLocalTransRotations",
]


@core.indexmod
def FillCoordTransRotations(
    frame,
    transform_store_key="OnlineRaDecRotation",
    end_coord_sys=MapCoordReference.Equatorial,
    do_bad_transform=False,
    bs_az_key="RawBoresightAz",
    bs_el_key="RawBoresightEl",
    bs_ra_key="OnlineBoresightRa",
    bs_dec_key="OnlineBoresightDec",
    offset_az_key="OffsetBoresightAz",
    offset_el_key="OffsetBoresightEl",
    offset_ra_key="OnlineOffsetRa",
    offset_dec_key="OnlineOffsetDec",
):
    """
    Calculates the rotation quaternions that take the point (1,0,0) (so az=el=0)
    in local coordinates to the coordinates specified by end_coord_sys and
    stores them in transform_store_key.  This encodes the boresight pointing and
    any rotations about this boresight pointing due to coordinate system
    changes, az/el bearing tilt, etc.

    Arguments
    ---------
    transform_store_key : string
        The key where the output transformation quaternion will be stored.
        If already present in the frame, this calculation will be skipped.
    end_coord_sys : MapCoordReference
        If Local, the transformation is computed using the negative of the
        detector delta angle.  Otherwise the detector angle is not inverted.
    do_bad_transform : bool
        If end_coord_sys is not Local and this argument is True, the offset
        keys are ignored and the coordinate transformation does not take
        into account rotation about the boresight.
    bs_az_key, bs_el_key : string
        Boresight coordinates in the local coordinate system.  If end_coord_sys
        is Local, only these two keys are required.
    bs_ra_key, bs_dec_key : string
        Boresight coordinates in the output coordinate system.  If
        do_bad_transform is True, only these two keys and the previous two keys
        are required.
    offset_az_key, offset_el_key, offset_ra_key, offset_dec_key : string
        Local and output coordinates computed at a small offset from boresight.
        These keys are required if do_bad_transform is False, and will be used
        to account for any rotation about boresight in the coordinate
        transformation.
    """
    if frame.type != core.G3FrameType.Scan:
        return

    if transform_store_key in frame:
        core.log_debug("Transform already computed, skipping")
        return

    trans = G3TimestreamQuat()
    if end_coord_sys == MapCoordReference.Local:
        trans = get_origin_rotator_timestream(
            frame[bs_az_key], frame[bs_el_key], end_coord_sys
        )
    else:
        if do_bad_transform:
            core.log_debug("You are doing the old calculation for pointing")
            trans = get_origin_rotator_timestream(
                frame[bs_ra_key], frame[bs_dec_key], end_coord_sys
            )
        else:
            trans = get_boresight_rotator_timestream(
                frame[bs_az_key],
                frame[bs_el_key],
                frame[bs_ra_key],
                frame[bs_dec_key],
                frame[offset_az_key],
                frame[offset_el_key],
                frame[offset_ra_key],
                frame[offset_dec_key],
            )
    frame[transform_store_key] = trans


@core.indexmod
def EquatorialToGalacticTransRotations(
    frame, eq_trans_key="OnlineRaDecRotation", out_key="OnlineGalacticRotation"
):
    """
    Takes a quaternion vector specifying the rotation to FK5 (Equatorial)
    boresight and converts it into a rotation to Galactic J2000 boresight
    coordinates.

    Use this to convert the output of FillCoordTransRotations to Galactic
    coordinates.
    """

    if frame.type != core.G3FrameType.Scan:
        return
    gal_trans = get_fk5_j2000_to_gal_quat() * frame[eq_trans_key]
    frame[out_key] = gal_trans


@core.indexmod
def AddLocalTransRotations(
    frame, az_key="RawBoresightAz", el_key="RawBoresightEl", out_key="RawAzElRotation"
):
    """
    Creates the transform for boresight pointing for az el based maps.  This is
    equivalent to FillCoordTransRotations with end_coord_sys in Local
    coordinates.

    Right now it's using the coordinate system where delta = -el because of
    implementation details.  All of the maps generated with this will be upside
    down in el.
    """

    if frame.type != core.G3FrameType.Scan:
        return
    local_trans = get_origin_rotator_timestream(
        frame[az_key], frame[el_key], MapCoordReference.Local
    )
    frame[out_key] = local_trans




#!/usr/bin/env python

from spt3g.core import G3Pipeline, G3Module
import unittest

class MySubModule(G3Module):
    def __init__(self, myarg=None):
        super(MySubModule, self).__init__()

    def Process(self, frame):
        pass

class MyModule:
    def __init__(self, myarg=None):
        pass
    def __call__(self, frame):
        pass

class TestPipelineModules(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pipe = G3Pipeline()

    def test_add_g3module(self):
        self.pipe.Add(MySubModule)
        self.pipe.Add(MySubModule, "argval")

    def test_add_generic_module(self):
        self.pipe.Add(MyModule)
        self.pipe.Add(MyModule, "argval")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from spt3g.core import G3Frame
import unittest

class TestFrames(unittest.TestCase):
    def test_put_get_delete(self):
        frame = G3Frame()
        self.assertEqual(list(frame.keys()), [])
        frame["key1"] = "val1"
        self.assertEqual(frame["key1"], "val1")
        del frame["key1"]
        with self.assertRaises(KeyError):
            frame["key1"]

    def test_iter(self):
        frame = G3Frame()
        self.assertEqual(list(frame.keys()), [])
        frame["key1"] = "val1"
        frame["key2"] = "val2"

        keys = []
        for key in frame:
            keys.append(key)
        self.assertEqual(len(keys), 2)
        self.assertTrue("key1" in keys)
        self.assertTrue("key2" in keys)

    def test_items(self):
        frame = G3Frame()
        self.assertEqual(list(frame.keys()), [])
        frame["key1"] = "val1"
        frame["key2"] = "val2"

        keys = []
        vals = []
        for k, v in frame.items():
            keys.append(k)
            vals.append(v)
        self.assertEqual(len(keys), 2)
        self.assertEqual(len(vals), 2)
        self.assertTrue("key1" in keys)
        self.assertTrue("key2" in keys)
        self.assertTrue("val1" in vals)
        self.assertTrue("val2" in vals)

    def test_values(self):
        frame = G3Frame()
        self.assertEqual(list(frame.keys()), [])
        frame["key1"] = "val1"
        frame["key2"] = "val2"
        keys = []
        vals = []
        self.assertTrue("val1" in frame.values())
        self.assertTrue("val2" in frame.values())
        del frame["key1"]
        del frame["key2"]
        self.assertTrue("val1" not in frame.values())

    def test_len(self):
        frame = G3Frame()
        self.assertEqual(len(frame), 0)
        frame["key1"] = "val1"
        frame["key2"] = "val2"
        self.assertEqual(len(frame), 2)
        del frame["key1"]
        del frame["key2"]
        self.assertEqual(len(frame), 0)


if __name__ == '__main__':
    unittest.main()



from . import G3Module, G3Pipeline, G3PipelineInfo, G3Frame, G3FrameType, G3Time, G3ModuleConfig
import types
import re
import textwrap


def g3decorator(func):
    """
    Mark argument as a decorator that can be found by automated
    documentation tools.
    """
    func.__g3decorator__ = True
    return func


@g3decorator
def usefulfunc(func):
    '''
    Mark argument as a useful function that can be found by automated
    documentation tools.

    Example
    -------
    ::

        @core.usefulfunc
        def do_some_science(data):
            science(data)
    '''

    func.__g3usefulfunc__ = True

    return func

# document thyself
usefulfunc.__g3usefulfunc__ = True


class PipeSegmentDocstring(type):
    @property
    def __doc__(cls):
        return '''
    Use as a decorator for a pre-assembled set of pipeline modules. Makes a
    pseudo-module consisting of several inputs. Use this to introspect the
    segment to find out what it does, or use :func:`pipesegment_nodoc` if your
    module does anything complicated.

    Example
    -------
    ::

        @core.pipesegment
        def standardfiltering(
            pipe,
            PolyOrder=4,
            MaskedHighPassEll=6000,
            Input='CalTimestreams',
            Output='FilteredTimestreams',
        ):
            pipe.Add(analysis.PolyFilter, PolyOrder=PolyOrder, Input=Input,
                Output='__Temp' + Output)
            pipe.Add(analysis.MaskedHighPass, MaskedHighPassEll=MaskedHighPassEll,
                Input='__Temp' + Output, Output=Output)
            def cleanup(frame):
                del frame['__Temp' + Output]
            pipe.Add(cleanup)

        pipe.Add(standardfiltering, PolyOrder=3)
    '''


@g3decorator
class pipesegment(metaclass=PipeSegmentDocstring):
    def __init__(self, func):
        self.func = self.__wrapped__ = func
        self.__pipesegment__ = func.__pipesegment__ = True

    def __call__(self, pipe, *args, **kwargs):
        return self.func(pipe, *args, **kwargs)

    @property
    def __doc__(self):
        """
        Create a dummy pipeline for introspection.  Generates a docstring when
        the __doc__ attribute is accessed.
        """
        if hasattr(self, "_autodoc"):
            return self._autodoc

        introdoc = textwrap.dedent(getattr(self.func, "__doc__", "")) or ""
        if introdoc:
            introdoc += "\n\n"
        introdoc += '\nEquivalent to\n-------------\n\n::\n\n'
        doclines = []
        class PotemkinPipe(object):
            def Add(self, thing, *args, **kwargs):
                if hasattr(thing, '__wrapped__'):
                    modname = thing.__wrapped__.__module__
                else:
                    modname = thing.__module__
                doc = 'pipe.Add(%s.%s' % (modname, thing.__name__)
                for arg in args:
                    doc += ', %s' % repr(arg)
                for arg in kwargs:
                    # remove object hashes
                    s = re.sub('<(.*) at (.*)>', '<\\1>', repr(kwargs[arg]))
                    doc += ', %s=%s' % (arg, s)
                doc += ')'
                doclines.append(doc)
        fake = PotemkinPipe()
        try:
            self.func(fake)
        except Exception as e:
            doclines.append('Exception evaluating equivalence (%s)' % (str(e), ))
        introdoc += '    ' + '\n    '.join(doclines) + '\n'

        self._autodoc = introdoc
        return self._autodoc

    @property
    def __name__(self):
        return self.func.__name__


@g3decorator
def pipesegment_nodoc(func):
    """
    Use as a decorator for a pre-assembled set of pipeline modules. Makes a
    pseudo-module consisting of several inputs.  Use this variant instead of
    :class:`pipesegment` to avoid introspection if your pipeline does anything
    complicated.

    Example
    -------
    ::

        @core.pipesegment_nodoc
        def standardfiltering(
            pipe,
            PolyOrder=4,
            MaskedHighPassEll=6000,
            Input='CalTimestreams',
            Output='FilteredTimestreams',
        ):
            pipe.Add(analysis.PolyFilter, PolyOrder=PolyOrder, Input=Input,
                Output='__Temp' + Output)
            pipe.Add(analysis.MaskedHighPass, MaskedHighPassEll=MaskedHighPassEll,
                Input='__Temp' + Output, Output=Output)
            def cleanup(frame):
                del frame['__Temp' + Output]
            pipe.Add(cleanup)

        pipe.Add(standardfiltering, PolyOrder=3)
    """
    func.__pipesegment__ = True
    return func


@g3decorator
def indexmod(func):
    '''
    Mark argument as a processing module that can be found by automated
    documentation tools.

    Example
    -------
    ::

        @core.indexmod
        def dostuff(frame):
            dosomestuff()
    '''

    func.__g3module__ = True

    return func

def build_pymodule(pycallable, *args, **kwargs):
    '''Convert a python callable and arguments into a core.G3Module by hook or by crook'''
    from .g3logging import log_fatal

    if isinstance(pycallable, G3Module):
        return pycallable

    if not callable(pycallable):
        log_fatal('Argument not a python callable', unit = 'G3Pipeline')

    if type(pycallable) == types.FunctionType:
        class PyFuncModule(G3Module):
            def Process(self, fr):
                return pycallable(fr, *args, **kwargs)
        return PyFuncModule()

    # If this is not a function, and it is callable, it is a class. If it is
    # a non-instantiated class, instantiate it.

    try:
        # issubclass() throws TypeError if argument is instantiated
        issubclass(pycallable, G3Module)
        isclass = True
    except TypeError:
        isclass = False

    # Instantiate if necessary. Not in try block so we don't miss TypeError
    if isclass:
        pycallable = pycallable(*args, **kwargs)
    else:
        # No way to handle arguments in this call; assert there are none
        if len(args) != 0 or len(kwargs) != 0:
            log_fatal('Cannot pass through arguments when passed instantiated class', unit = 'G3Pipeline')

    # See if it was a Python G3Module subclass
    if isinstance(pycallable, G3Module):
        return pycallable

    # This is a python callable that is not a module, so wrap it
    class PyCallObjModule(G3Module):
        def Process(self, fr):
            return pycallable(fr)

    return PyCallObjModule()

class _add_pipeline_info(G3Module):
    def __init__(self):
        from .. import version
        import socket, getpass

        G3Module.__init__(self)
        self.infoemitted = False
        self.buffer = []

        self.pipelineinfo = G3PipelineInfo()
        self.pipelineinfo.vcs_url = version.upstream_url
        self.pipelineinfo.vcs_branch = version.upstream_branch
        self.pipelineinfo.vcs_revision = version.revision
        self.pipelineinfo.vcs_localdiffs = version.localdiffs
        self.pipelineinfo.vcs_versionname = version.versionname
        self.pipelineinfo.vcs_fullversion = version.fullversion
        self.pipelineinfo.vcs_githash = version.gitrevision

        self.pipelineinfo.hostname = socket.gethostname()
        self.pipelineinfo.user = getpass.getuser()
    def Process(self, fr):
        if self.infoemitted:
            if fr.type == G3FrameType.PipelineInfo and \
              self.originalpi is not None and \
              self.originalpi == list(fr.keys()):
                # Deduplicate PipelineInfo frames identical to one that we
                # added to earlier, which avoids false semi-duplicates down
                # the line when processing multiple files.
                return False
            return fr

        # Allow limited reordering of metadata
        if fr.type in [G3FrameType.Observation, G3FrameType.Wiring,
          G3FrameType.Calibration]:
            self.buffer.append(fr)
            return []

        self.infoemitted = True
        if fr.type == G3FrameType.PipelineInfo:
            self.originalpi = list(fr.keys())
            fr[str(G3Time.Now())] = self.pipelineinfo
            self.buffer.append(fr)
        else:
            self.originalpi = None
            f = G3Frame(G3FrameType.PipelineInfo)
            f[str(G3Time.Now())] = self.pipelineinfo
            self.buffer += [f, fr]
        rv = self.buffer
        del self.buffer
        return rv

def PipelineAddCallable(self, callable, name=None, subprocess=False, **kwargs):
    '''
    Add a processing module to the pipeline. It can be any subclass of
    spt3g.core.G3Module or any Python callable, either an instance or
    a class. Positional and keyword arguments are passed through to the
    argument's constructor (if a class) or as additional arguments to a
    function. If subprocess is set to True, this module will be
    run in a separate process.
    '''

    addpipelineinfo = False
    if not hasattr(self, '_pipelineinfo'):
        self._pipelineinfo = _add_pipeline_info()
        addpipelineinfo = True

    if not hasattr(self, 'nameprefix'):
        self.nameprefix = ''
    if (hasattr(callable, '__name__')):
        callable_name = callable.__name__
    elif (hasattr(callable, '__class__')):
        callable_name = callable.__class__.__name__
    else:
        raise RuntimeError("Cannot establish name of pipeline module")
    if name is None:
        name = '%s.%s' % (callable.__module__, callable_name)
    name = self.nameprefix + name

    # Record module configuration for root objects
    if self.nameprefix == '': 
        modconfig = G3ModuleConfig()
        modconfig.instancename = name
        modconfig.modname = '%s.%s' % (callable.__module__, callable_name)
        for k,v in kwargs.items():
            tostore = v
            try:
                if v.npix_allocated > 0:
                    # Don't store full sky maps as configuration options. It
                    # just wastes a ton of disk space with simulations.
                    tostore = v.clone(False)
            except:
                # If that threw an exception, it either isn't a map or dropping
                # data didn't work, so just don't bother.
                pass
            modconfig[k] = tostore
        self._pipelineinfo.pipelineinfo.modules.append(modconfig)

    # Deal with the segment case
    if hasattr(callable, '__pipesegment__'):
        # Prefix module names added by segments with the segment name
        oldnameprefix = self.nameprefix
        self.nameprefix = name + '/'

        rv = callable(self, **kwargs)

        self.nameprefix = oldnameprefix

    else:
        # Otherwise it's a module
        pymod = build_pymodule(callable, **kwargs)
        if subprocess:
            from .multiprocess import Subproc
            pymod = build_pymodule(Subproc(pymod, name=name))

        rv = self._Add_(pymod, name=name)

    if addpipelineinfo:
        self._Add_(self._pipelineinfo, name='_pipelineinfo')

    return rv

# Add this as G3Pipeline's Add method so it takes any Python callable
G3Pipeline.Add = PipelineAddCallable
G3Pipeline.__repr__ = lambda self: repr(self._pipelineinfo.pipelineinfo) if hasattr(self, '_pipelineinfo') else 'pipe = {}.G3Pipeline()'.format(G3Pipeline.__module__)




from .. import core
from . import DfMuxHousekeepingMap, HkBoardInfo, HkMezzanineInfo, HkModuleInfo, HkChannelInfo
from .TuberClient import TuberClient
import socket, struct, time

@core.indexmod
class LegacyHousekeepingConsumer(object):
    '''
    Collect housekeeping data from the legacy (i.e. SPTpol-ish) mux boards
    defined in the wiring map. Will add a key called 'DfMuxHousekeeping' to
    any housekeeping frame that goes by containing the data as of the arrival
    of the housekeeping frame. Use in conjunction with a
    dfmux.PeriodicHousekeepingCollector to get data at fixed intervals.
   
    If collecting real-time data, you may want to set subprocess=True when
    adding this module.
    '''
    def __init__(self):
        self.tuber = None
    def __call__(self, frame):
        if frame.type == core.G3FrameType.Wiring:
            wmap = frame['WiringMap']
            board_ips = set([m.board_ip for m in wmap.values()])

            self.tuber = [(ip, TuberClient(socket.inet_ntoa(struct.pack('i', m.board_ip)))) for ip in board_ips]

        if frame.type == core.G3FrameType.Housekeeping:
            for board,board_tuber in self.tuber:
                board_tuber.CallMethod('Dfmux', ['classic_dfmux_dump', 'classic_housekeeping'], False)

            hkdata = DfMuxHousekeepingMap()
            for board,board_tuber in self.tuber:
                dat = dict(board_tuber.GetReply()[0]['result'])
                dat.update(board_tuber.GetReply()[1]['result'])

                serial = struct.unpack(">i", struct.pack("i", board))[0] & 0xff
                hkdata[serial] = self.HousekeepingFromJSON(dat)
            frame['DfMuxHousekeeping'] = hkdata

    @classmethod
    def HousekeepingFromJSON(cls, dat):
        '''
        Build HKBoardInfo object from the union of the JSON blobs
        returned by the classic_housekeeping and class_dfmux_dump
        Tuber calls
        '''
        # Board-global quantities
        boardhk = HkBoardInfo()

        if dat['ts']['port'] == 'IRIG test':
            boardhk.timestamp = core.G3Time(dat['ts']['s']*core.G3Units.s)
        else:
            year = dat['ts']['y']
            if year == 0:
                # It probably isn't 1900
                systime = time.gmtime()
                # Check for New Year's, assuming no more than 24 hours clock slew
                if dat['ts']['d'] == 1 and systime.tm_yday >= 365:
                    year = systime.tm_year + 1
                elif dat['ts']['d'] >= 365 and systime.tm_yday == 1:
                    year = systime.tm_year - 1
                else:
                    year = systime.tm_year
                year -= 2000
            boardhk.timestamp = core.G3Time(y=year,d=dat['ts']['d'],h=dat['ts']['h'],m=dat['ts']['m'],s=dat['ts']['s'],ss=dat['ts']['ss'])

        boardhk.timestamp_port = str(dat['ts']['port'])
        boardhk.serial = 'N/A'
        boardhk.fir_stage = dat['fir_stage']
        for i in dat['voltages']['mb'].items():
            boardhk.voltages[str(i[0])] = i[1]
        for i in dat['temperatures']['mb'].items():
            boardhk.temperatures[str(i[0])] = i[1]

        # Mezzanines
        for mezz in [1,2]:
            mezzhk = HkMezzanineInfo()
            mezzhk.power = dat['voltages']['mezz%d' % mezz]['v3p'] > 0
            mezzhk.present = True
            mezzhk.serial = 'N/A'
            mezzhk.part_number = 'SPTpol Mezz'
            mezzhk.revision = 'N/A'
            for i in dat['voltages']['mezz%d' % mezz].items():
                if isinstance(i[1], float):
                    mezzhk.voltages[str(i[0])] = i[1]

            # Modules
            for mod in [1,2]:
                wire = (mezz - 1)*2 + mod

                # XXX: SQUID data (bias point, FLL) not in HK call!
                # XXX: SQUID tuning state
                modhk = HkModuleInfo()
                modhk.carrier_gain = dat['mezz_gains']['wire%d' % wire]['carrier']
                modhk.nuller_gain = dat['mezz_gains']['wire%d' % wire]['nuller']
                modhk.demod_gain = dat['mezz_gains']['wire%d' % wire]['demod']

                modhk.carrier_railed = dat['voltages']['mezz%d' % mezz]['overload_dmfs_car_%s' % (['a', 'b'][mod-1])]
                modhk.nuller_railed = dat['voltages']['mezz%d' % mezz]['overload_dmfs_nul_%s' % (['a', 'b'][mod-1])]
                modhk.demod_railed = dat['voltages']['mezz%d' % mezz]['overload_dmfd_%s' % (['a', 'b'][mod-1])]

                modhk.module_number = mod

                for chan in range(1, dat['config']['dmfd_channels_per_wire'] + 1):
                    chanhk = HkChannelInfo()
                    chanhk.channel_number = chan
                    chanhk.dan_gain = dat['dan']['gain'][wire-1][chan-1]
                    chanhk.dan_accumulator_enable = dat['dan']['accumulator_enable'][wire-1][chan-1]
                    chanhk.dan_feedback_enable = dat['dan']['feedback_enable'][wire-1][chan-1]
                    chanhk.dan_streaming_enable = dat['dan']['streaming_enable'][wire-1][chan-1]
                    chanhk.dan_railed = dat['dan']['accumulator_railed'][wire-1][chan-1]

                    # Channel params: (freq, phase, amp)
                    carrier = dat['dmfs']['carrier'][wire-1][chan-1]
                    nuller = dat['dmfs']['nuller'][wire-1][chan-1]
                    # Demod params: (freq, phase)
                    demod = dat['dmfd']['demod'][wire-1][chan-1]
                    chanhk.carrier_amplitude = carrier[2]/(2.**23 - 1.)
                    chanhk.nuller_amplitude = nuller[2]/(2.**23 - 1.)
                    chanhk.carrier_frequency = carrier[0]*25.e6/(2.**32-1)*core.G3Units.Hz
                    chanhk.demod_frequency = demod[0]*25.e6/(2.**32-1)*core.G3Units.Hz

                    modhk.channels[chan] = chanhk
                mezzhk.modules[mod] = modhk
            boardhk.mezz[mezz] = mezzhk
    
        return boardhk

    @classmethod
    def HousekeepingForBoard(cls, hostname):
        '''
        Return HkBoardInfo object from a board. Useful for debugging.
        '''
        t = TuberClient(hostname)
        jsonblob = t.CallMethod('Dfmux', ['classic_dfmux_dump', 'classic_housekeeping'])
        data = dict(jsonblob[0]['result'])
        data.update(jsonblob[1]['result'])

        return cls.HousekeepingFromJSON(data)




#!/usr/bin/env python

import numpy as np
from spt3g import core, maps

# Test initialization from sparse arrays

a = np.arange(1500,dtype='float')
a[0] = -1
b = np.arange(1500,dtype='int')
x = maps.HealpixSkyMap(b, a, 64, True, False, maps.MapCoordReference.Equatorial)
x.shift_ra = False
assert(x.nside == 64)
assert(x.npix_allocated == 1500)
assert(x[1499] == 1499)
assert(x[1501] == 0)

# check attributes
attrs = ["nside", "res", "nested", "shift_ra", "units", "pol_type", "pol_conv", "coord_ref", "weighted"]
xc = x.clone()
for attr in attrs:
    assert getattr(xc, attr) == getattr(x, attr), "Attribute {} mismatched on clone()".format(attr)
xc = x.clone(False)
for attr in attrs:
    assert getattr(xc, attr) == getattr(x, attr), "Attribute {} mismatched on clone(False)".format(attr)

# Test conversion between representations is lossless
# Along the way, test nonzero_pixels() and rebin() once for each
# representation.

x.dense = True # Indexed-sparse to dense
assert(x.npix_allocated == x.size)
assert(x[1499] == 1499)
assert(x[1501] == 0)

x.ringsparse = True # Dense to ring-sparse
assert(x.nside == 64)
assert(x.npix_allocated == 1500)
for i in range(1,1500):
    assert(x[i] == i)
assert(x[1501] == 0)

x.indexedsparse = True # Ring to indexed
assert(x.nside == 64)
assert(x.npix_allocated == 1500)
for i in range(1,1500):
    assert(x[i] == i)
assert(x[1501] == 0)

k,v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a)) # Lazy test that doesn't care about order

import healpy as hp
p0 = hp.nest2ring(64, (hp.ring2nest(32, 0) * 4 + np.arange(4)).astype(int))
v0 = sum([x[int(i)] for i in p0])

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(np.sum(x2) == np.sum(v))

x.shift_ra = True

# check attributes
xc = x.clone()
for attr in attrs:
    assert getattr(xc, attr) == getattr(x, attr), "Attribute {} mismatched on clone()".format(attr)
xc = x.clone(False)
for attr in attrs:
    assert getattr(xc, attr) == getattr(x, attr), "Attribute {} mismatched on clone(False)".format(attr)

x.ringsparse = True # Indexed to ring
assert(x.nside == 64)
assert(x.npix_allocated == 1512) # shifted ringsparse is less efficient
assert(x[1499] == 1499)
assert(x[1501] == 0)

k,v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a))

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(np.sum(x2) == np.sum(v))

x.dense = True # Ring to dense
x.shift_ra = False
assert(x[1499] == 1499)
assert(x.npix_allocated == x.size)
assert(len({x[i] for i in range(x.size) if x[i] != 0}) == 1500)
assert(x[1501] == 0)

k,v = x.nonzero_pixels()
assert(len(k) == len(v) == 1500)
assert(set(v) == set(a))

x2 = x.rebin(2, norm=False)
assert(x2[0] == v0)
assert(np.sum(x2) == np.sum(v))

x.indexedsparse = True # Dense to indexed
assert(x.nside == 64)
assert(x.npix_allocated == 1500)
for i in range(1,1500):
    assert(x[i] == i)
assert(x[1501] == 0)

# Initialization from dense arrays
a = np.arange(49152)
x = maps.HealpixSkyMap(a)

for i in range(0, len(a)):
    assert(x[i] == i)

assert((np.asarray(x) == a).all())

# Conersion to ring-sparse again (trickiest, this makes sure we get all rings)
x.ringsparse = True
assert(x.npix_allocated == len(a) - 1) # First element was zero
for i in range(0, len(a)):
    assert(x[i] == i)

assert((np.asarray(x) == a).all())
assert(x.dense) # Should be dense again

# test ra shifting
x.shift_ra = False
x.ringsparse = True
ki, vi = x.nonzero_pixels()
ii = np.argsort(ki)
ki = np.asarray(ki)[ii]
vi = np.asarray(vi)[ii]

x.shift_ra = True
kr, vr = x.nonzero_pixels()
ii = np.argsort(kr)
kr = np.asarray(kr)[ii]
vr = np.asarray(vr)[ii]
assert((ki == kr).all())
assert((vi == vr).all())

x.shift_ra = False
ki, vi = x.nonzero_pixels()
ii = np.argsort(ki)
ki = np.asarray(ki)[ii]
vi = np.asarray(vi)[ii]
assert((ki == kr).all())
assert((vi == vr).all())

# angles
pixels = np.arange(x.size, dtype=int)
alpha, delta = x.pixels_to_angles(pixels)
alpha, delta = np.asarray(alpha), np.asarray(delta)
alpha[alpha < 0] += 360 * core.G3Units.deg
theta, phi = hp.pix2ang(x.nside, pixels)

assert(np.allclose(alpha, phi))
assert(np.allclose(delta, np.pi / 2 - theta))

pixels2 = x.angles_to_pixels(alpha, delta)
assert(np.allclose(pixels, pixels2))

# interpolation
dx = x.res / 2.0
vx = np.asarray(x.get_interp_values(alpha + dx, delta + dx))
vh = hp.get_interp_val(np.asarray(x), theta - dx, phi + dx)
assert(not np.allclose(vx, vh))

# Query disc
avec = []
dvec = []
rvec = []
masked = set()
alpha = 30 * core.G3Units.deg
radius = 5 * core.G3Units.deg
for delta in range(-90, 91, 10):
    delta *= core.G3Units.deg
    theta = 90 * core.G3Units.deg - delta
    avec.append(alpha)
    dvec.append(delta)
    rvec.append(radius)
    pix1 = np.asarray(x.query_disc(alpha, delta, radius), dtype=int)
    v = hp.ang2vec(theta, alpha)
    pix2 = np.asarray(hp.query_disc(x.nside, v, radius), dtype=int)
    masked |= set(pix2)
    assert not (set(pix1) ^ set(pix2))

mask = maps.make_point_source_mask(x, np.asarray(avec), np.asarray(dvec), np.asarray(rvec))
assert not (set(mask.nonzero()) ^ masked)

# Conversion to/from flatsky maps
fm_stub = maps.FlatSkyMap(
    300, 300, core.G3Units.arcmin, proj=maps.MapProjection.ProjZEA
)
fm = maps.maputils.healpix_to_flatsky(x, map_stub=fm_stub)
x2 = maps.maputils.flatsky_to_healpix(fm, map_stub=x.clone(False))

hitpix = np.asarray(x2) > 0
assert(np.allclose(np.asarray(x)[hitpix], np.asarray(x2)[hitpix]))


# Coordinate system rotations
a = np.arange(49152)
x = maps.HealpixSkyMap(a)
alpha, delta = maps.get_ra_dec_map(x)

# equatorial to galactic
xgal = x.clone(False)
x.coord_ref = maps.MapCoordReference.Equatorial
xgal.coord_ref = maps.MapCoordReference.Galactic
maps.reproj_map(x, xgal)
ra, dec = maps.azel.convert_gal_to_radec(alpha, delta)
pix = xgal.angles_to_pixels(ra, dec)
assert(np.allclose(np.asarray(pix), np.asarray(xgal)))

# ... and back
xeq = x.clone(False)
x.coord_ref = maps.MapCoordReference.Galactic
xeq.coord_ref = maps.MapCoordReference.Equatorial
maps.reproj_map(x, xeq)
lon, lat = maps.azel.convert_radec_to_gal(alpha, delta)
pix = xeq.angles_to_pixels(lon, lat)
assert(np.allclose(np.asarray(pix), np.asarray(xeq)))

# high-res
hx = maps.HealpixSkyMap(16384)
pix1 = hx.angle_to_pixel(0, -45 * core.G3Units.deg)
pix2 = hp.ang2pix(16384, 3 * np.pi / 4, 0)
assert(pix1 == pix2)



#!/usr/bin/env python

from spt3g import core, dfmux
import socket, argparse
import numpy as np

parser = argparse.ArgumentParser(description='Record dfmux data to a NetCDF file', prog='ledgerman')
parser.add_argument('hardware_map', metavar='/path/to/hwm.yaml', help='Path to hardware map YAML file')
parser.add_argument('output', metavar='output.nc', help='Path to output NetCDF file')

parser.add_argument('-v', dest='verbose', action='store_true',
                    help='Verbose mode (print all frames)')
parser.add_argument('-u', dest='udp', action='store_true',
                    help='Use multicast UDP for data collection instead of '
                    'SCTP for compatibility with old firmware.')
parser.add_argument('-t', dest='threads', type=int, default=None,
                    help='Number of listener threads to use (SCTP only)')
parser.add_argument('-a', dest='align', action='store_true',
                    help='Align sampling. This has to happen once, but'
                    ' will break any existing DAN loops when run.')
parser.add_argument('-s', dest='system_time', action='store_true',
                    help='Replace board time with system time when data'
                    ' received. Useful if your board is in IRIG_TEST mode.')
parser.add_argument('-P', dest='physnames', action='store_true',
                    help='Use physical bolometer names rather than channel names')
parser.add_argument('-p', dest='pathstring', action='store', default=[], nargs='+',
                    help='Only record channels that match channel path string,'
                    ' e.g. 005/5/2/3/* saves data for all bolometers on crate'
                    ' 005, slot 5, mezzanine 2, module 3.')
parser.add_argument('-b', dest='state', action='store', default=[], nargs='+',
                    help='Only record bolometers that have a state that'
                    ' matches the supplied string(s), e.g. overbiased tuned')
args = parser.parse_args()


# Import pydfmux later since it can take a while
import pydfmux

core.log_notice('Initializing hardware map and boards', unit='Ledgerman')
hwm = pydfmux.load_session(open(args.hardware_map, 'r'))['hardware_map']
if hwm.query(pydfmux.IceCrate).count() > 0:
    hwm.query(pydfmux.IceCrate).resolve()

# make sure that the hardware map is consistent with what's on the IceBoards
if args.state:
    hwm.query(pydfmux.Bolometer).load_bolo_states()

if args.align:
    core.log_notice('Aligning board sampling, this will break any existing DAN loops!',
                    unit='Ledgerman')
    hwm.query(pydfmux.IceBoard).set_fir_stage(6)
    hwm.query(pydfmux.IceBoard).align_sampling()

core.log_notice('Beginning data acquisition', unit='Ledgerman')

# get board serial numbers only for the channels that we are recording
if args.pathstring:
    chan_map_query = hwm.channel_maps_from_pstring(args.pathstring)
else:
    chan_map_query = hwm.query(pydfmux.ChannelMapping)
if args.state:
    chan_map_query = chan_map_query.join(pydfmux.ChannelMapping, pydfmux.Bolometer).filter(pydfmux.Bolometer.state._in(args.state))
serial_list = np.unique(np.array([cm.iceboard.serial for cm in chan_map_query]))

# Set up DfMux consumer
pipe = core.G3Pipeline()
builder = dfmux.DfMuxBuilder([int(serial) for serial in serial_list])

if args.udp:
    # Get the local IP(s) to use to connect to the boards by opening test
    # connections. Using a set rather than a list deduplicates the results.
    local_ips = set()
    for board in hwm.query(pydfmux.core.dfmux.IceBoard):
        testsock = socket.create_connection(('iceboard' + board.serial + '.local', 80))
        local_ips.add(testsock.getsockname()[0])
        testsock.close()
    core.log_notice('Creating listeners for %d boards on interfaces: %s' % (hwm.query(pydfmux.core.dfmux.IceBoard).count(), ', '.join(local_ips)), unit='Ledgerman')

    # Build mapping dictionary for old (64x) firmware
    v2_mapping = {'iceboard' + str(serial) + '.local': int(serial) for serial in serial_list}

    # Set up listeners per network segment and point them at the event builder
    collectors = [dfmux.DfMuxCollector(ip, builder, v2_mapping) for ip in local_ips]
else:
    hosts = ['iceboard' + str(serial) + '.local' for serial in serial_list]

    if args.threads is None:
        args.threads = len(hosts)
    if args.threads > len(hosts):
        args.threads = len(hosts)
    collectors = [dfmux.DfMuxCollector(builder, hosts[i::args.threads])
                  for i in range(args.threads)]

pipe.Add(builder)

# Insert current hardware map into data stream. This is critical to get the
# board ID -> IP mapping needed to do anything useful with the data
pipe.Add(dfmux.PyDfMuxHardwareMapInjector, pydfmux_hwm=hwm, pathstring=args.pathstring, state=args.state)

if args.physnames:
    from spt3g import calibration
    # Swap the normal bolometer names for their physical_name equivalents,
    # which can make laboratory tasks simpler -- we don't care about
    # long-term data archiving here.
    pipe.Add(dfmux.PyDfMuxBolometerPropertiesInjector, pydfmux_hwm=hwm)

    class SwapPhysLogicalNames(object):
        def __init__(self):
            self.calframe = None
            self.wiringframe = None
            self.sent = False

        def __call__(self, frame):
            if self.sent:
                return
            if frame.type == core.G3FrameType.Wiring:
                self.wiringframe = frame
                return []
            if frame.type == core.G3FrameType.Calibration:
                self.calframe = frame
                return []
            if self.calframe is None or self.wiringframe is None:
                raise Exception('Data before wiring and cal frames!')
            w = self.wiringframe['WiringMap']
            del self.wiringframe['WiringMap']
            c = self.calframe['NominalBolometerProperties']

            new_wiring = dfmux.DfMuxWiringMap()
            for k, d in w.iteritems():
                new_wiring[c[k].physical_name] = d
            self.wiringframe['WiringMap'] = new_wiring

            self.sent = True
            return [self.wiringframe, frame]
    pipe.Add(SwapPhysLogicalNames)

if args.system_time:
    def sub_system_time(frame):
        if frame.type != core.G3FrameType.Timepoint:
            return
        del frame['EventHeader']
        frame['EventHeader'] = core.G3Time.Now()
    pipe.Add(sub_system_time)

if args.verbose:
    pipe.Add(core.Dump)


pipe.Add(dfmux.NetCDFDump, filename=args.output)

for collector in collectors:
    collector.Start()
pipe.Run()




from .. import core
from . import DfMuxMetaSample
import numpy as np

def get_empty_timepoint(sample_time):
    tp_frame = core.G3Frame(core.G3FrameType.Timepoint)
    tp_frame["EventHeader"] = sample_time
    tp_frame["DfMux"] = DfMuxMetaSample()
    return tp_frame

class AddMissingTimepoints(object):
    def __init__(self, expected_sample_rate = (20e6 * core.G3Units.Hz)/ 2**17):
        self.samp_spacing_ = 1.0 / expected_sample_rate
        self.prev_sample_time_ = None

    def __call__(self, frame):
        if frame.type != core.G3FrameType.Timepoint:
            return            
        if 'EventHeader' not in frame:
            return
        new_time = frame['EventHeader']
        if self.prev_sample_time_ == None:
            self.prev_sample_time_ = new_time
            return
        t_delt = float(new_time - self.prev_sample_time_)
        n_missing_samples = int(round(t_delt / samp_spacing_ )) - 1
        samp_times = [new_time - samp_spacing_ * i for i in range(1, 1+n_missing_samples)]
        new_packets = [get_empty_timepoint(st) for st in samp_times]
        new_packets.append(frame)
        return new_packets

@core.indexmod
class FillMissingTimepointFrames(object):
    '''
    Search for Timepoint frames that come at larger-than-expected intervals.
    Fill in the missin frames with garbage, and mark them as garbage for
    conversion to NaNs later.
    '''
    def __init__(self):
        self.dt = None
        self.last_frame = None
        self.buffer = []

    def __call__(self, fr):
        if fr.type != core.G3FrameType.Timepoint:
            return
        if self.dt is None:
            # Calculate the sample rate from the first 20 frames
            self.buffer.append(fr)
            if len(self.buffer) < 20:
                return []
            else:
                times = [_fr['EventHeader'].time for _fr in self.buffer]
                intervals = np.diff(times)
                self.dt = min(intervals[intervals > 0])
                # Now process the frames in the buffer
                out = sum([self._find_and_fill_frames(_fr) for _fr in self.buffer], [])
                self.buffer = []
                return out
        return self._find_and_fill_frames(fr)

    def _find_and_fill_frames(self, fr):
        # This function does most of the work:
        # Find frames with separation that is too large, and fill the gaps
        # with new frames.
        if self.last_frame is None:
            self.last_frame = fr
            return [fr]
        this_dt = fr['EventHeader'].time - self.last_frame['EventHeader'].time
        if this_dt > 1.5 * self.dt:
            # We're missing at least one frame
            nframes = int(np.round(this_dt / self.dt) - 1)
            injected_dt = this_dt / (nframes + 1)
            injected_frs = [core.G3Frame(core.G3FrameType.Timepoint)
                            for i in range(nframes)]
            for i, new_fr in enumerate(injected_frs):
                new_fr['EventHeader'] = self.last_frame['EventHeader'] + \
                                     injected_dt * (i + 1)
                new_fr['DfMux'] = fr['DfMux'].copy()
                try:
                    new_fr['CalibratorOn'] = fr['CalibratorOn']
                except KeyError:
                    pass
                new_fr['GarbageData'] = True
            self.last_frame = fr
            return injected_frs + [fr]
        self.last_frame = fr
        return [fr]




import numpy as np
from . import Quat, G3VectorQuat, G3TimestreamQuat

__all__ = []

quat_types = (Quat, G3VectorQuat, G3TimestreamQuat)


def quat_ufunc(self, ufunc, method, *inputs, **kwargs):
    """Numpy ufunc interface for vectorized quaternion methods."""
    if ufunc.__name__ in ["isinf", "isnan", "isfinite"] and len(inputs) == 1:
        v = getattr(ufunc, method)(np.asarray(inputs[0]), **kwargs)
        if ufunc.__name__ == "isfinite":
            return np.all(v, axis=-1)
        return np.any(v, axis=-1)
    if ufunc.__name__.startswith("logical"):
        args = []
        for arg in inputs:
            if isinstance(arg, quat_types):
                arg = np.any(np.asarray(arg), axis=-1)
            args.append(arg)
        return getattr(ufunc, method)(*args, **kwargs)
    if method != "__call__" or kwargs:
        return NotImplemented
    if len(inputs) == 1:
        if ufunc.__name__ == "absolute":
            return self.abs()
        if ufunc.__name__ == "negative":
            return self.__neg__()
        if ufunc.__name__ == "conjugate":
            return self.conj()
        if ufunc.__name__ == "reciprocal":
            return Quat(1, 0, 0, 0) / self
    if len(inputs) == 2 and np.isscalar(inputs[1]):
        if ufunc.__name__ == "power":
            return self.__pow__(inputs[1])
    return NotImplemented


Quat.__array_ufunc__ = quat_ufunc
G3VectorQuat.__array_ufunc__ = quat_ufunc
G3TimestreamQuat.__array_ufunc__ = quat_ufunc



import numpy
from spt3g import core, dfmux

pipe = core.G3Pipeline()
builder = dfmux.DfMuxBuilder(1)
collector = dfmux.DfMuxCollector("192.168.1.4", builder)

pipe.Add(builder)

startpoint = {}
nframes = 0

def dump(frame):
	global nframes
	#print('Frame: %s' % str(frame['EventHeader']))
	nframes += 1
	for i in frame['DfMux']:
		ip = i.key()
		data = i.data()
		for module in data:
			if module.key() not in startpoint:
				startpoint[module.key()] = {}
			samps = numpy.abs(module.data())
			for chan in range(0,16):
				if chan not in startpoint[module.key()]:
					startpoint[module.key()][chan] = 0
				if nframes < 200:
					startpoint[module.key()][chan] = max(startpoint[module.key()][chan], max(samps[chan*2], samps[chan*2 + 1]))
				rail = startpoint[module.key()][chan]*3
				if samps[chan*2] > rail or samps[chan*2 + 1] > rail:
					print('Channel %d hosed (I %d, Q %d)' % (chan, samps[chan*2], samps[chan*2 + 1]))


pipe.Add(dump)
#pipe.Add(core.G3Writer('dfmux.g3'))

collector.Start()
pipe.Run()
